import json
import textwrap
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Literal, TypeAlias

import click
import dagster_shared.check as check
from dagster_dg_core.context import DgContext
from dagster_dg_core.utils import snakecase
from dagster_shared.scaffold import scaffold_subtree
from dagster_shared.serdes.objects.package_entry import EnvRegistryKey
from dagster_shared.seven import match_module_pattern

ScaffoldFormatOptions: TypeAlias = Literal["yaml", "python"]


def scaffold_component(
    *, dg_context: DgContext, class_name: str, module_name: str, model: bool
) -> None:
    module_parts = module_name.split(".")
    module_path = dg_context.root_path
    for i in range(len(dg_context.root_module_name.split(".")), len(module_parts) - 1):
        module = ".".join(module_parts[: i + 1])
        module_path = dg_context.get_path_for_local_module(module, require_exists=False)
        if not module_path.exists():
            click.echo(f"Creating module at: {module_path}")
            module_path.mkdir()
        (module_path / "__init__.py").touch()

    scaffold_subtree(
        path=module_path,
        name_placeholder="COMPONENT_TYPE_NAME_PLACEHOLDER",
        project_template_path=Path(__file__).parent / "templates" / "COMPONENT_TYPE",
        project_name=module_parts[-1],
        name=class_name,
        model=model,
    )

    # backcompat -- dagster_dg_cli.plugin entry point
    if dg_context.has_registry_module_entry_point:
        with open(module_path / "__init__.py") as f:
            lines = f.readlines()
        lines.append(f"from {module_name} import {class_name} as {class_name}\n")
        with open(module_path / "__init__.py", "w") as f:
            f.writelines(lines)

    # no plugin entry point, add to project plugin modules
    else:
        project_config = check.not_none(dg_context.config.project)
        if not any(
            match_module_pattern(module_name, pattern)
            for pattern in project_config.registry_modules
        ):
            dg_context.add_project_registry_module(module_name)
    click.echo(f"Scaffolded Dagster component at {module_path}/{module_parts[-1]}.py.")


# ########################
# ##### INLINE COMPONENT
# ########################


def scaffold_inline_component(
    path: Path,
    typename: str,
    superclass: str | None,
    dg_context: "DgContext",
) -> None:
    full_path = dg_context.defs_path / path
    full_path.mkdir(parents=True, exist_ok=True)
    click.echo(
        f"Creating a Dagster inline component and corresponding component instance at {path}."
    )

    component_path = full_path / f"{snakecase(typename)}.py"
    if superclass:
        key = EnvRegistryKey.from_typename(superclass)
        superclass_import_lines = [
            f"from {key.namespace} import {key.name}",
        ]
        superclass_list = key.name
    else:
        superclass_import_lines = []
        superclass_list = "dg.Component, dg.Model, dg.Resolvable"

    component_lines = [
        "import dagster as dg",
        *superclass_import_lines,
        "",
        f"class {typename}({superclass_list}):",
        "    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:",
        "        return dg.Definitions()",
    ]
    component_path.write_text("\n".join(component_lines))

    containing_module = ".".join(
        [
            dg_context.defs_module_name,
            str(path).replace("/", "."),
            f"{snakecase(typename)}",
        ]
    )
    defs_path = full_path / "defs.yaml"
    defs_path.write_text(
        textwrap.dedent(f"""
        type: {containing_module}.{typename}
        attributes: {{}}
    """).strip()
    )


# ####################
# ##### REGISTRY OBJECT
# ####################


def scaffold_registry_object(
    path: Path,
    typename: str,
    scaffold_params: Mapping[str, Any] | None,
    dg_context: "DgContext",
    scaffold_format: ScaffoldFormatOptions,
    append: bool = False,
) -> None:
    from dagster.components.component_scaffolding import scaffold_object

    scaffold_object(
        path,
        typename,
        json.dumps(scaffold_params) if scaffold_params else None,
        scaffold_format,
        dg_context.root_path,
        append,
    )



"""Asset models for REST-like API."""

from pydantic import BaseModel


class DgApiAssetHealthMetadata(BaseModel):
    """Asset health metadata details."""

    failed_run_id: str | None
    num_failed_partitions: int | None
    num_missing_partitions: int | None
    total_num_partitions: int | None
    num_failed_checks: int | None
    num_warning_checks: int | None
    total_num_checks: int | None
    last_materialized_timestamp: float | None


class DgApiAssetMaterialization(BaseModel):
    """Latest asset materialization information."""

    timestamp: float | None
    run_id: str | None
    partition: str | None


class DgApiAssetFreshnessInfo(BaseModel):
    """Asset freshness information."""

    current_lag_minutes: float | None
    current_minutes_late: float | None
    latest_materialization_minutes_late: float | None
    maximum_lag_minutes: float | None
    cron_schedule: str | None


class DgApiAssetChecksStatus(BaseModel):
    """Asset checks execution status."""

    status: str | None  # HEALTHY, WARNING, DEGRADED, UNKNOWN, NOT_APPLICABLE
    num_failed_checks: int | None
    num_warning_checks: int | None
    total_num_checks: int | None


class DgApiAssetStatus(BaseModel):
    """Asset status information for status view."""

    asset_health: str | None  # Overall health status
    materialization_status: str | None
    freshness_status: str | None
    asset_checks_status: str | None
    health_metadata: DgApiAssetHealthMetadata | None
    latest_materialization: DgApiAssetMaterialization | None
    freshness_info: DgApiAssetFreshnessInfo | None
    checks_status: DgApiAssetChecksStatus | None


class DgApiAsset(BaseModel):
    """Asset resource model."""

    id: str
    asset_key: str  # "my/asset/key"
    asset_key_parts: list[str]  # ["my", "asset", "key"]
    description: str | None
    group_name: str
    kinds: list[str]
    metadata_entries: list[dict]
    dependency_keys: list[str] = []  # ["dep/one", "dep/two"]
    # Status fields - populated only for status view
    status: DgApiAssetStatus | None = None

    class Config:
        from_attributes = True


class DgApiAssetList(BaseModel):
    """GET /api/assets response."""

    items: list[DgApiAsset]
    cursor: str | None  # Next cursor for pagination
    has_more: bool  # Whether more results exist



"""GraphQL implementation for schedule operations."""

from typing import TYPE_CHECKING, Any

from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.schedule import DgApiSchedule, DgApiScheduleList

LIST_SCHEDULES_QUERY = """
query ListSchedules($repositorySelector: RepositorySelector!) {
    schedulesOrError(repositorySelector: $repositorySelector) {
        __typename
        ... on Schedules {
            results {
                id
                name
                cronSchedule
                pipelineName
                description
                executionTimezone
                scheduleState {
                    status
                }
            }
        }
        ... on RepositoryNotFoundError {
            message
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""

LIST_REPOSITORIES_QUERY = """
query ListRepositories {
    repositoriesOrError {
        __typename
        ... on RepositoryConnection {
            nodes {
                name
                location {
                    name
                }
                schedules {
                    id
                    name
                    cronSchedule
                    pipelineName
                    description
                    executionTimezone
                    scheduleState {
                        status
                    }
                }
            }
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""

GET_SCHEDULE_QUERY = """
query GetSchedule($scheduleSelector: ScheduleSelector!) {
    scheduleOrError(scheduleSelector: $scheduleSelector) {
        __typename
        ... on Schedule {
            id
            name
            cronSchedule
            pipelineName
            description
            executionTimezone
            scheduleState {
                status
            }
        }
        ... on ScheduleNotFoundError {
            message
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


def process_repositories_response(graphql_response: dict[str, Any]) -> "DgApiScheduleList":
    """Process GraphQL response from repositories query into DgApiScheduleList."""
    from dagster_dg_cli.api_layer.schemas.schedule import (
        DgApiSchedule,
        DgApiScheduleList,
        DgApiScheduleStatus,
    )

    repositories_result = graphql_response.get("repositoriesOrError")
    if not repositories_result:
        raise Exception("No repositories data in GraphQL response")

    typename = repositories_result.get("__typename")
    if typename != "RepositoryConnection":
        error_msg = repositories_result.get("message", f"GraphQL error: {typename}")
        raise Exception(error_msg)

    repositories_data = repositories_result.get("nodes", [])

    schedules = []
    for repo in repositories_data:
        repo_name = repo.get("name", "")
        location_name = repo.get("location", {}).get("name", "")
        code_location_origin = (
            f"{location_name}@{repo_name}" if location_name and repo_name else None
        )

        for s in repo.get("schedules", []):
            schedule_state = s.get("scheduleState", {})
            status = schedule_state.get("status", "STOPPED")

            schedule = DgApiSchedule(
                id=s["id"],
                name=s["name"],
                status=DgApiScheduleStatus(status),
                cron_schedule=s.get("cronSchedule", ""),
                pipeline_name=s.get("pipelineName", ""),
                description=s.get("description"),
                execution_timezone=s.get("executionTimezone"),
                code_location_origin=code_location_origin,
                next_tick_timestamp=None,
            )
            schedules.append(schedule)

    return DgApiScheduleList(
        items=schedules,
        total=len(schedules),
    )


def process_schedules_response(graphql_response: dict[str, Any]) -> "DgApiScheduleList":
    """Process GraphQL response into DgApiScheduleList.

    Args:
        graphql_response: Raw GraphQL response containing "schedulesOrError"

    Returns:
        DgApiScheduleList: Processed schedule data
    """
    from dagster_dg_cli.api_layer.schemas.schedule import (
        DgApiSchedule,
        DgApiScheduleList,
        DgApiScheduleStatus,
    )

    schedules_result = graphql_response.get("schedulesOrError")
    if not schedules_result:
        raise Exception("No schedules data in GraphQL response")

    typename = schedules_result.get("__typename")
    if typename != "Schedules":
        error_msg = schedules_result.get("message", f"GraphQL error: {typename}")
        raise Exception(error_msg)

    schedules_data = schedules_result.get("results", [])

    schedules = []
    for s in schedules_data:
        schedule_state = s.get("scheduleState", {})
        status = schedule_state.get("status", "STOPPED")

        schedule = DgApiSchedule(
            id=s["id"],
            name=s["name"],
            status=DgApiScheduleStatus(status),
            cron_schedule=s.get("cronSchedule", ""),
            pipeline_name=s.get("pipelineName", ""),
            description=s.get("description"),
            execution_timezone=s.get("executionTimezone"),
            code_location_origin=None,
            next_tick_timestamp=None,
        )
        schedules.append(schedule)

    return DgApiScheduleList(
        items=schedules,
        total=len(schedules),
    )


def process_schedule_response(graphql_response: dict[str, Any]) -> "DgApiSchedule":
    """Process GraphQL response into DgApiSchedule.

    Args:
        graphql_response: Raw GraphQL response containing "scheduleOrError"

    Returns:
        DgApiSchedule: Processed schedule data
    """
    from dagster_dg_cli.api_layer.schemas.schedule import DgApiSchedule, DgApiScheduleStatus

    schedule_result = graphql_response.get("scheduleOrError")
    if not schedule_result:
        raise Exception("No schedule data in GraphQL response")

    typename = schedule_result.get("__typename")
    if typename != "Schedule":
        error_msg = schedule_result.get("message", f"GraphQL error: {typename}")
        raise Exception(error_msg)

    schedule_state = schedule_result.get("scheduleState", {})
    status = schedule_state.get("status", "STOPPED")

    return DgApiSchedule(
        id=schedule_result["id"],
        name=schedule_result["name"],
        status=DgApiScheduleStatus(status),
        cron_schedule=schedule_result.get("cronSchedule", ""),
        pipeline_name=schedule_result.get("pipelineName", ""),
        description=schedule_result.get("description"),
        execution_timezone=schedule_result.get("executionTimezone"),
        code_location_origin=None,
        next_tick_timestamp=None,
    )


def list_schedules_via_graphql(
    client: IGraphQLClient,
    repository_location_name: str | None = None,
    repository_name: str | None = None,
) -> "DgApiScheduleList":
    """Fetch schedules using GraphQL."""
    if repository_location_name and repository_name:
        variables = {
            "repositorySelector": {
                "repositoryLocationName": repository_location_name,
                "repositoryName": repository_name,
            }
        }
        result = client.execute(LIST_SCHEDULES_QUERY, variables)
        return process_schedules_response(result)
    else:
        result = client.execute(LIST_REPOSITORIES_QUERY)
        return process_repositories_response(result)


def get_schedule_via_graphql(
    client: IGraphQLClient,
    schedule_name: str,
    repository_location_name: str,
    repository_name: str,
) -> "DgApiSchedule":
    """Get single schedule via GraphQL."""
    variables = {
        "scheduleSelector": {
            "scheduleName": schedule_name,
            "repositoryLocationName": repository_location_name,
            "repositoryName": repository_name,
        }
    }

    try:
        result = client.execute(GET_SCHEDULE_QUERY, variables)
        if (
            result.get("data", {}).get("scheduleOrError", {}).get("__typename")
            == "ScheduleNotFoundError"
        ):
            raise Exception(f"Schedule not found: {schedule_name}")
        return process_schedule_response(result)
    except Exception:
        raise


def get_schedule_by_name_via_graphql(client: IGraphQLClient, schedule_name: str) -> "DgApiSchedule":
    """Get schedule by name, searching across all code locations."""
    result = client.execute(LIST_REPOSITORIES_QUERY)
    all_schedules = process_repositories_response(result)

    matching_schedules = [
        schedule for schedule in all_schedules.items if schedule.name == schedule_name
    ]

    if not matching_schedules:
        raise Exception(f"Schedule not found: {schedule_name}")

    if len(matching_schedules) > 1:
        code_location_origins = [
            schedule.code_location_origin or "unknown" for schedule in matching_schedules
        ]
        raise Exception(
            f"Multiple schedules found with name '{schedule_name}' in code locations: {', '.join(code_location_origins)}"
        )

    return matching_schedules[0]



"""GraphQL adapter for run metadata."""

from dagster_dg_cli.api_layer.schemas.run import DgApiRun, DgApiRunStatus
from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

RUN_METADATA_QUERY = """
query DgApiRunMetadataQuery($runId: ID!) {
    runOrError(runId: $runId) {
        __typename
        ... on Run {
            id
            runId
            status
            creationTime
            startTime
            endTime
            jobName
        }
        ... on RunNotFoundError {
            message
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


def get_run_via_graphql(client: IGraphQLClient, run_id: str) -> DgApiRun:
    """Get run metadata via GraphQL."""
    variables = {"runId": run_id}

    # Try/catch here is a temporary solution until we fix the GQL-- client.execute is throwing an
    # error if the run is not found, instead of returning the error in the response.
    try:
        result = client.execute(RUN_METADATA_QUERY, variables)
    except Exception:
        raise Exception(f"Run not found: {run_id}")

    run_result = result.get("runOrError")
    if not run_result:
        raise Exception(f"Run not found: {run_id}")

    typename = run_result.get("__typename")

    # Handle GraphQL errors
    if run_result.get("__typename") != "Run":
        error_msg = run_result.get("message", f"Graphql error: {typename}")
        raise Exception(error_msg)

    return DgApiRun(
        id=run_result["runId"],
        status=DgApiRunStatus(run_result["status"]),
        created_at=run_result["creationTime"],
        started_at=run_result.get("startTime"),
        ended_at=run_result.get("endTime"),
        job_name=run_result.get("jobName"),
    )



"""Output formatters for CLI display."""

import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.agent import DgApiAgent, DgApiAgentList
    from dagster_dg_cli.api_layer.schemas.asset import DgApiAsset, DgApiAssetList
    from dagster_dg_cli.api_layer.schemas.deployment import Deployment, DeploymentList
    from dagster_dg_cli.api_layer.schemas.secret import DgApiSecret, DgApiSecretList


def format_deployments(deployments: "DeploymentList", as_json: bool) -> str:
    """Format deployment list for output."""
    if as_json:
        return deployments.model_dump_json(indent=2)

    lines = []
    for deployment in deployments.items:
        lines.extend(
            [
                f"Name: {deployment.name}",
                f"ID: {deployment.id}",
                f"Type: {deployment.type.value}",
                "",  # Empty line between deployments
            ]
        )

    return "\n".join(lines).rstrip()  # Remove trailing empty line


def format_deployment(deployment: "Deployment", as_json: bool) -> str:
    """Format single deployment for output."""
    if as_json:
        return deployment.model_dump_json(indent=2)

    lines = [
        f"Name: {deployment.name}",
        f"ID: {deployment.id}",
        f"Type: {deployment.type.value}",
    ]

    return "\n".join(lines)


def format_assets(assets: "DgApiAssetList", as_json: bool) -> str:
    """Format asset list for output."""
    if as_json:
        return assets.model_dump_json(indent=2)

    lines = []
    for asset in assets.items:
        asset_lines = [
            f"Asset Key: {asset.asset_key}",
            f"ID: {asset.id}",
            f"Description: {asset.description or 'None'}",
            f"Group: {asset.group_name}",
            f"Kinds: {', '.join(asset.kinds) if asset.kinds else 'None'}",
            f"Deps: {', '.join(asset.dependency_keys) if asset.dependency_keys else 'None'}",
        ]

        # Add status information if present
        if asset.status:
            asset_lines.extend(_format_asset_status_lines(asset.status))

        asset_lines.append("")  # Empty line between assets
        lines.extend(asset_lines)

    return "\n".join(lines).rstrip()  # Remove trailing empty line


def _format_timestamp(timestamp: float, unit: str = "seconds") -> str:
    """Format timestamp for human-readable display.

    Args:
        timestamp: The timestamp value
        unit: Either "milliseconds" or "seconds" to indicate the timestamp unit
    """
    try:
        if unit == "milliseconds":
            dt = datetime.datetime.fromtimestamp(timestamp / 1000)
        elif unit == "seconds":
            dt = datetime.datetime.fromtimestamp(timestamp)
        else:
            raise ValueError(f"Unsupported unit: {unit}")
        return dt.strftime("%Y-%m-%d %H:%M:%S")
    except (ValueError, OSError):
        return f"Invalid timestamp: {timestamp}"


def _format_asset_status_lines(status) -> list[str]:
    """Format asset status information into human-readable lines."""
    lines = []

    # Overall health status
    if status.asset_health:
        lines.append(f"Asset Health: {status.asset_health}")
    if status.materialization_status:
        lines.append(f"Materialization Status: {status.materialization_status}")
    if status.freshness_status:
        lines.append(f"Freshness Status: {status.freshness_status}")
    if status.asset_checks_status:
        lines.append(f"Asset Checks Status: {status.asset_checks_status}")

    # Health metadata details
    if status.health_metadata:
        metadata = status.health_metadata
        if metadata.failed_run_id:
            lines.append(f"Failed Run ID: {metadata.failed_run_id}")
        if metadata.num_failed_partitions is not None:
            lines.append(
                f"Failed Partitions: {metadata.num_failed_partitions}/{metadata.total_num_partitions or 'unknown'}"
            )
        if metadata.num_failed_checks is not None:
            lines.append(
                f"Failed Checks: {metadata.num_failed_checks}/{metadata.total_num_checks or 'unknown'}"
            )
        if metadata.num_warning_checks is not None:
            lines.append(f"Warning Checks: {metadata.num_warning_checks}")
        if metadata.last_materialized_timestamp:
            lines.append(
                f"Last Materialized: {_format_timestamp(metadata.last_materialized_timestamp, 'milliseconds')}"
            )

    # Latest materialization
    if status.latest_materialization:
        mat = status.latest_materialization
        if mat.timestamp:
            lines.append(
                f"Latest Materialization: {_format_timestamp(mat.timestamp, 'milliseconds')}"
            )
        if mat.run_id:
            lines.append(f"Latest Run ID: {mat.run_id}")
        if mat.partition:
            lines.append(f"Latest Partition: {mat.partition}")

    # Freshness info
    if status.freshness_info:
        freshness = status.freshness_info
        if freshness.current_lag_minutes is not None:
            lines.append(f"Current Lag: {freshness.current_lag_minutes:.1f} minutes")
        if freshness.current_minutes_late is not None:
            lines.append(f"Minutes Late: {freshness.current_minutes_late:.1f}")
        if freshness.maximum_lag_minutes is not None:
            lines.append(f"Max Allowed Lag: {freshness.maximum_lag_minutes:.1f} minutes")
        if freshness.cron_schedule:
            lines.append(f"Freshness Schedule: {freshness.cron_schedule}")

    # Asset checks details
    if status.checks_status and status.checks_status.total_num_checks is not None:
        checks = status.checks_status
        lines.append(f"Total Checks: {checks.total_num_checks}")
        if checks.num_failed_checks is not None:
            lines.append(f"Failed Checks: {checks.num_failed_checks}")
        if checks.num_warning_checks is not None:
            lines.append(f"Warning Checks: {checks.num_warning_checks}")

    return lines


def format_asset(asset: "DgApiAsset", as_json: bool) -> str:
    """Format single asset for output."""
    if as_json:
        return asset.model_dump_json(indent=2)

    lines = [
        f"Asset Key: {asset.asset_key}",
        f"ID: {asset.id}",
        f"Description: {asset.description or 'None'}",
        f"Group: {asset.group_name}",
        f"Kinds: {', '.join(asset.kinds) if asset.kinds else 'None'}",
        f"Deps: {', '.join(asset.dependency_keys) if asset.dependency_keys else 'None'}",
    ]

    # Add status information if present
    if asset.status:
        lines.append("")
        lines.append("Status Information:")
        status_lines = _format_asset_status_lines(asset.status)
        for line in status_lines:
            lines.append(f"  {line}")

    if asset.metadata_entries:
        lines.append("")
        lines.append("Metadata:")
        for entry in asset.metadata_entries:
            value = entry.get("description", "")
            for key in ["text", "url", "path", "jsonString", "mdStr"]:
                if entry.get(key):
                    value = entry[key]
                    break
            for key in ["floatValue", "intValue", "boolValue"]:
                if entry.get(key) is not None:
                    value = str(entry[key])
                    break
            lines.append(f"  {entry['label']}: {value}")

    return "\n".join(lines)


def format_agents(agents: "DgApiAgentList", as_json: bool) -> str:
    """Format agent list for output."""
    if as_json:
        return agents.model_dump_json(indent=2)

    lines = []
    for agent in agents.items:
        # Use agent_label if available, otherwise format as "Agent {first_8_chars_of_id}"
        display_label = agent.agent_label or f"Agent {agent.id[:8]}"
        lines.extend(
            [
                f"Label: {display_label}",
                f"ID: {agent.id}",
                f"Status: {agent.status.value}",
                f"Last Heartbeat: {_format_timestamp(agent.last_heartbeat_time, 'seconds') if agent.last_heartbeat_time else 'Never'}",
                "",  # Empty line between agents
            ]
        )

    return "\n".join(lines).rstrip()  # Remove trailing empty line


def format_agent(agent: "DgApiAgent", as_json: bool) -> str:
    """Format single agent for output."""
    if as_json:
        return agent.model_dump_json(indent=2)

    # Use agent_label if available, otherwise format as "Agent {first_8_chars_of_id}"
    display_label = agent.agent_label or f"Agent {agent.id[:8]}"
    lines = [
        f"Label: {display_label}",
        f"ID: {agent.id}",
        f"Status: {agent.status.value}",
        f"Last Heartbeat: {_format_timestamp(agent.last_heartbeat_time, 'seconds') if agent.last_heartbeat_time else 'Never'}",
    ]

    if agent.metadata:
        lines.append("")
        lines.append("Metadata:")
        for meta in agent.metadata:
            lines.append(f"  {meta.key}: {meta.value}")

    return "\n".join(lines)


def format_secrets(secrets: "DgApiSecretList", as_json: bool) -> str:
    """Format secret list for output.

    Note: Secret values are never shown in list format for security.
    """
    if as_json:
        return secrets.model_dump_json(indent=2)

    if not secrets.items:
        return "No secrets found."

    lines = []
    for secret in secrets.items:
        lines.extend(
            [
                f"Name: {secret.name}",
                f"Locations: {', '.join(secret.location_names) if secret.location_names else 'All code locations'}",
                f"Scopes: {_format_secret_scopes(secret)}",
                f"Can Edit: {'Yes' if secret.can_edit_secret else 'No'}",
                f"Can View Value: {'Yes' if secret.can_view_secret_value else 'No'}",
            ]
        )

        if secret.updated_by:
            lines.append(f"Updated By: {secret.updated_by.email}")

        if secret.update_timestamp:
            lines.append(f"Updated: {secret.update_timestamp.strftime('%Y-%m-%d %H:%M:%S')}")

        lines.append("")  # Empty line between secrets

    return "\n".join(lines).rstrip()  # Remove trailing empty line


def format_secret(secret: "DgApiSecret", as_json: bool, show_value: bool = False) -> str:
    """Format single secret for output.

    Args:
        secret: Secret to format
        as_json: Whether to output JSON format
        show_value: Whether to include the secret value (security sensitive)
    """
    import json

    if as_json:
        if not show_value and secret.value is not None:
            # Create a copy with hidden value for JSON output
            secret_dict = secret.model_dump()
            secret_dict["value"] = "<hidden>"
            return json.dumps(secret_dict, indent=2, default=str)
        return secret.model_dump_json(indent=2)

    lines = [
        f"Name: {secret.name}",
        f"Locations: {', '.join(secret.location_names) if secret.location_names else 'All code locations'}",
        f"Scopes: {_format_secret_scopes(secret)}",
        "Permissions:",
        f"  Can Edit: {'Yes' if secret.can_edit_secret else 'No'}",
        f"  Can View Value: {'Yes' if secret.can_view_secret_value else 'No'}",
    ]

    # Show value only if explicitly requested and available
    if show_value:
        if secret.value is not None:
            lines.extend(
                [
                    "",
                    "Value:",
                    f"  {secret.value}",
                ]
            )
        else:
            lines.extend(
                [
                    "",
                    "Value: <not available - you may not have permission to view this value>",
                ]
            )
    else:
        lines.extend(
            [
                "",
                "Value: <hidden - use --show-value to display>",
            ]
        )

    if secret.updated_by:
        lines.extend(
            [
                "",
                f"Updated By: {secret.updated_by.email}",
            ]
        )

    if secret.update_timestamp:
        lines.append(f"Updated: {secret.update_timestamp.strftime('%Y-%m-%d %H:%M:%S')}")

    return "\n".join(lines)


def _format_secret_scopes(secret: "DgApiSecret") -> str:
    """Format secret scopes into human-readable string."""
    scopes = []

    if secret.full_deployment_scope:
        scopes.append("Full Deployment")

    if secret.all_branch_deployments_scope:
        scopes.append("All Branch Deployments")

    if secret.specific_branch_deployment_scope:
        scopes.append(f"Branch: {secret.specific_branch_deployment_scope}")

    if secret.local_deployment_scope:
        scopes.append("Local Deployment")

    return ", ".join(scopes) if scopes else "None"



import contextlib
import tempfile
from collections.abc import Generator
from pathlib import Path
from typing import NamedTuple
from unittest import mock
from unittest.mock import patch

import pytest
import responses
import yaml
from dagster_cloud_cli.commands.ci import BuildStrategy
from dagster_cloud_cli.core.pex_builder.deps import BuildMethod
from dagster_cloud_cli.types import SnapshotBaseDeploymentCondition
from dagster_dg_cli.cli.plus.deploy import DEFAULT_STATEDIR_PATH
from dagster_dg_core.utils import pushd
from dagster_shared.plus.config import DagsterPlusCliConfig
from dagster_test.dg_utils.utils import (
    ProxyRunner,
    assert_runner_result,
    isolated_example_project_foo_bar,
)

from dagster_dg_cli_tests.cli_tests.plus_tests.utils import (
    PYTHON_VERSION,
    mock_hybrid_response,
    mock_serverless_response,
)


@pytest.fixture
def logged_in_dg_cli_config(empty_dg_cli_config):
    config = DagsterPlusCliConfig(
        organization="hooli",
        user_token="fake-user-token",
        default_deployment="prod",
    )
    config.write()
    yield


@pytest.fixture
def empty_dg_cli_config(monkeypatch):
    with (
        tempfile.TemporaryDirectory() as tmp_dg_dir,
    ):
        config_path = Path(tmp_dg_dir) / "dg.toml"
        monkeypatch.setenv("DG_CLI_CONFIG", config_path)
        # Unset DG_USE_EDITABLE_DAGSTER to mimic CI environment where it's not set
        monkeypatch.delenv("DG_USE_EDITABLE_DAGSTER", raising=False)

        # Also patch the function that checks the environment variable to return False
        from dagster_dg_core.shared_options import EDITABLE_DAGSTER_OPTIONS

        monkeypatch.setattr(
            "dagster_dg_core.shared_options.is_use_editable_env_var_true", lambda: False
        )

        # Patch the actual option's default value
        editable_option = EDITABLE_DAGSTER_OPTIONS["use_editable_dagster"]
        editable_option.default = None

        config = DagsterPlusCliConfig(
            organization="",
            user_token="",
            default_deployment="",
        )
        config.write()
        yield config_path


@pytest.fixture
def build_yaml_file(project):
    build_yaml_path = "build.yaml"
    try:
        with open(build_yaml_path, "w") as f:
            f.write("registry: my-repo\ndirectory: .")
        yield build_yaml_path
    finally:
        Path(build_yaml_path).unlink()


@pytest.fixture
def container_context_yaml_file(project):
    yaml_path = "container_context.yaml"
    try:
        with open(yaml_path, "w") as f:
            f.write("""k8s:
  env_secrets:
    - project-secret
""")
        yield yaml_path
    finally:
        Path(yaml_path).unlink()


@pytest.fixture
def workspace_build_yaml_file(workspace):
    build_yaml_path = workspace / "build.yaml"
    try:
        with open(build_yaml_path, "w") as f:
            f.write("registry: my-workspace-repo\ndirectory: .")
        yield build_yaml_path
    finally:
        Path(build_yaml_path).unlink()


@pytest.fixture
def workspace_container_context_yaml_file(workspace):
    yaml_path = workspace / "container_context.yaml"
    try:
        with open(yaml_path, "w") as f:
            f.write("""k8s:
  env_secrets:
    - workspace-secret
""")

        yield yaml_path
    finally:
        Path(yaml_path).unlink()


@pytest.fixture
def workspace_project_build_yaml_file(workspace):
    build_yaml_path = workspace / "foo-bar" / "build.yaml"
    try:
        with open(build_yaml_path, "w") as f:
            f.write("registry: my-project-repo\ndirectory: .")
        yield build_yaml_path
    finally:
        Path(build_yaml_path).unlink()


@pytest.fixture
def workspace_project_container_context_yaml_file(workspace):
    yaml_path = workspace / "foo-bar" / "container_context.yaml"
    try:
        with open(yaml_path, "w") as f:
            f.write("""k8s:
  env_secrets:
    - project-secret
""")
        yield yaml_path
    finally:
        Path(yaml_path).unlink()


@pytest.fixture(scope="module")
def runner():
    with ProxyRunner.test(use_fixed_test_components=True) as the_runner:
        yield the_runner


# make this a single project in a workspace and change the path instead


@pytest.fixture(scope="module")
def project_setup_path(runner):
    with isolated_example_project_foo_bar(
        runner, use_editable_dagster=False, in_workspace=True
    ) as project_path:
        with pushd(project_path.parent):
            args = [
                "project",
                "foo-bar-2",
                "--uv-sync",
            ]

            result = runner.invoke_create_dagster(*args)
            assert result.exit_code == 0
            assert_runner_result(result)

        yield project_path


def _cleanup_dockerfiles(workspace_path):
    if (workspace_path / "foo-bar" / "Dockerfile").exists():
        (workspace_path / "foo-bar" / "Dockerfile").unlink()
    if (workspace_path / "foo-bar-2" / "Dockerfile").exists():
        (workspace_path / "foo-bar-2" / "Dockerfile").unlink()


@pytest.fixture
def workspace(project_setup_path):
    with pushd(project_setup_path.parent):
        try:
            yield project_setup_path.parent
        finally:
            _cleanup_dockerfiles(project_setup_path.parent)


@pytest.fixture
def project(project_setup_path):
    with pushd(project_setup_path):
        try:
            yield project_setup_path
        finally:
            _cleanup_dockerfiles(project_setup_path.parent)


class MockedCloudCliCommands(NamedTuple):
    init: mock.MagicMock
    build: mock.MagicMock
    deploy: mock.MagicMock
    set_build_output: mock.MagicMock

    def reset_mocks(self):
        self.init.reset_mock()
        self.build.reset_mock()
        self.deploy.reset_mock()
        self.set_build_output.reset_mock()


@contextlib.contextmanager
def mock_external_dagster_cloud_cli_command() -> Generator[MockedCloudCliCommands, None, None]:
    with (
        patch(
            "dagster_cloud_cli.commands.ci.init_impl",
        ) as mock_init_command,
        patch(
            "dagster_cloud_cli.commands.ci.build_impl",
        ) as mock_build_command,
        patch(
            "dagster_cloud_cli.commands.ci.deploy_impl",
        ) as mock_deploy_command,
        patch(
            "dagster_cloud_cli.commands.ci.set_build_output",
        ) as mock_set_build_output_command,
        patch(
            "dagster_dg_cli.cli.plus.deploy.validation.validate_deploy_configuration",
        ) as mock_validation,
    ):
        # Ensure deploy_impl mock returns successfully without checking build state
        mock_deploy_command.return_value = None
        # Skip validation in tests
        mock_validation.return_value = None
        yield MockedCloudCliCommands(
            init=mock_init_command,
            build=mock_build_command,
            deploy=mock_deploy_command,
            set_build_output=mock_set_build_output_command,
        )


@responses.activate
def test_plus_deploy_command_agent_type_from_graphql(
    logged_in_dg_cli_config, project: Path, runner
):
    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        mock_serverless_response()

        result = runner.invoke("plus", "deploy", "--yes")
        assert result.exit_code == 0, result.output + " : " + str(result.exception)
        assert "No Dockerfile found - scaffolding a default one" in result.output

        mocked_cloud_cli_commands.build.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            build_directory=str(project.resolve()),
            dockerfile_path=str((project / "Dockerfile").absolute()),
            use_editable_dagster=False,
            build_strategy=BuildStrategy.docker,
            docker_image_tag=None,
            docker_base_image=None,
            docker_env=[],
            python_version=PYTHON_VERSION,
            pex_build_method=BuildMethod.DOCKER_FALLBACK,
            pex_deps_cache_from=None,
            pex_deps_cache_to=None,
            pex_base_image_tag=None,
            location_name=["foo-bar"],
        )

        mocked_cloud_cli_commands.reset_mocks()

        mock_serverless_response()

        result = runner.invoke("plus", "deploy", "build-and-push")
        assert not result.exit_code, result.output
        mocked_cloud_cli_commands.build.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            build_directory=str(project.resolve()),
            dockerfile_path=str((project / "Dockerfile").absolute()),
            use_editable_dagster=False,
            build_strategy=BuildStrategy.docker,
            docker_image_tag=None,
            docker_base_image=None,
            docker_env=[],
            python_version=PYTHON_VERSION,
            pex_build_method=BuildMethod.DOCKER_FALLBACK,
            pex_deps_cache_from=None,
            pex_deps_cache_to=None,
            pex_base_image_tag=None,
            location_name=["foo-bar"],
        )


def test_plus_deploy_command_serverless(logged_in_dg_cli_config, project: Path, runner):
    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert result.exit_code == 0, result.output + " : " + str(result.exception)
        assert "No Dockerfile found - scaffolding a default one" in result.output

        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(project.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            require_branch_deployment=False,
            git_url=None,
            dagster_env=None,
            location_name=[],
            snapshot_base_condition=None,
            status_url=None,
        )
        mocked_cloud_cli_commands.build.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            build_directory=str(project.resolve()),
            dockerfile_path=str((project / "Dockerfile").absolute()),
            use_editable_dagster=False,
            build_strategy=BuildStrategy.docker,
            docker_image_tag=None,
            docker_base_image=None,
            docker_env=[],
            python_version=PYTHON_VERSION,
            pex_build_method=BuildMethod.DOCKER_FALLBACK,
            pex_deps_cache_from=None,
            pex_deps_cache_to=None,
            pex_base_image_tag=None,
            location_name=["foo-bar"],
        )
        mocked_cloud_cli_commands.deploy.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            agent_heartbeat_timeout=mock.ANY,
            location_load_timeout=mock.ANY,
            location_name=[],
        )

        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert "Building using Dockerfile at" in result.output
        assert result.exit_code == 0, result.output + " : " + str(result.exception)


def test_plus_deploy_command_serverless_workspace(logged_in_dg_cli_config, workspace, runner):
    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        from dagster_cloud_cli.commands.ci import BuildStrategy

        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert result.exit_code == 0, result.output + " : " + str(result.exception)
        assert "No Dockerfile found - scaffolding a default one" in result.output

        workspace_project_path = workspace / "foo-bar"
        workspace_other_project_path = workspace / "foo-bar-2"

        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(workspace.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            require_branch_deployment=False,
            git_url=None,
            location_name=[],
            dagster_env=None,
            snapshot_base_condition=None,
            status_url=None,
        )
        mocked_cloud_cli_commands.build.assert_has_calls(
            [
                mock.call(
                    statedir=DEFAULT_STATEDIR_PATH,
                    build_directory=str(workspace_project_path.resolve()),
                    dockerfile_path=str((workspace_project_path / "Dockerfile").resolve()),
                    use_editable_dagster=False,
                    build_strategy=BuildStrategy.docker,
                    docker_image_tag=None,
                    docker_base_image=None,
                    docker_env=[],
                    python_version=PYTHON_VERSION,
                    pex_build_method=BuildMethod.DOCKER_FALLBACK,
                    pex_deps_cache_from=None,
                    pex_deps_cache_to=None,
                    pex_base_image_tag=None,
                    location_name=["foo-bar"],
                ),
                mock.call(
                    statedir=DEFAULT_STATEDIR_PATH,
                    build_directory=str(workspace_other_project_path.resolve()),
                    dockerfile_path=str((workspace_other_project_path / "Dockerfile").resolve()),
                    use_editable_dagster=False,
                    build_strategy=BuildStrategy.docker,
                    docker_image_tag=None,
                    docker_base_image=None,
                    docker_env=[],
                    python_version=PYTHON_VERSION,
                    pex_build_method=BuildMethod.DOCKER_FALLBACK,
                    pex_deps_cache_from=None,
                    pex_deps_cache_to=None,
                    pex_base_image_tag=None,
                    location_name=["foo-bar-2"],
                ),
            ]
        )

        mocked_cloud_cli_commands.deploy.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            agent_heartbeat_timeout=mock.ANY,
            location_load_timeout=mock.ANY,
            location_name=[],
        )

        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert "Building using Dockerfile at" in result.output
        assert result.exit_code == 0, result.output + " : " + str(result.exception)


def test_plus_deploy_command_invalid_location(logged_in_dg_cli_config, workspace, runner):
    with mock_external_dagster_cloud_cli_command():
        result = runner.invoke(
            "plus",
            "deploy",
            "--agent-type",
            "serverless",
            "--location-name",
            "doenotexist",
            "--yes",
        )
        assert result.exit_code != 0
        assert "The following requested locations do not exist: doenotexist" in result.output


def test_plus_deploy_command_valid_location(logged_in_dg_cli_config, workspace, runner):
    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        from dagster_cloud_cli.commands.ci import BuildStrategy

        result = runner.invoke(
            "plus",
            "deploy",
            "--agent-type",
            "serverless",
            "--location-name",
            "foo-bar-2",
            "--yes",
        )

        assert result.exit_code == 0, result.output + " : " + str(result.exception)
        assert "No Dockerfile found - scaffolding a default one" in result.output

        workspace_project_path = workspace / "foo-bar-2"

        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(workspace.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            require_branch_deployment=False,
            git_url=None,
            location_name=["foo-bar-2"],
            dagster_env=None,
            snapshot_base_condition=None,
            status_url=None,
        )
        mocked_cloud_cli_commands.build.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            build_directory=str(workspace_project_path.resolve()),
            dockerfile_path=str((workspace_project_path / "Dockerfile").resolve()),
            use_editable_dagster=False,
            build_strategy=BuildStrategy.docker,
            docker_image_tag=None,
            docker_base_image=None,
            docker_env=[],
            python_version=PYTHON_VERSION,
            pex_build_method=BuildMethod.DOCKER_FALLBACK,
            pex_deps_cache_from=None,
            pex_deps_cache_to=None,
            pex_base_image_tag=None,
            location_name=["foo-bar-2"],
        )
        mocked_cloud_cli_commands.deploy.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            agent_heartbeat_timeout=mock.ANY,
            location_load_timeout=mock.ANY,
            location_name=["foo-bar-2"],
        )

        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert "Building using Dockerfile at" in result.output
        assert result.exit_code == 0, result.output + " : " + str(result.exception)


def test_plus_deploy_command_no_login(empty_dg_cli_config, runner, project):
    result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
    assert result.exit_code != 0
    assert "Organization not specified" in result.output


def test_plus_deploy_on_branch(logged_in_dg_cli_config, project, runner, mocker):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="my-branch",
    )
    with mock_external_dagster_cloud_cli_command():
        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert result.exit_code == 0
        assert (
            "Deploying to the branch deployment for my-branch, with prod as the base deployment"
            in result.output
        )


def test_plus_deploy_on_branch_with_snapshot_base_condition(
    logged_in_dg_cli_config, project, runner, mocker
):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="my-branch",
    )
    with mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands:
        result = runner.invoke(
            "plus",
            "deploy",
            "--agent-type",
            "serverless",
            "--yes",
            "--snapshot-base-condition",
            "on-update",
            "--status-url",
            "https://my-status-url.com",
        )
        assert result.exit_code == 0
        assert (
            "Deploying to the branch deployment for my-branch, with prod as the base deployment"
            in result.output
        )
        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(project.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            require_branch_deployment=True,
            git_url=None,
            location_name=[],
            dagster_env=None,
            snapshot_base_condition=SnapshotBaseDeploymentCondition.ON_UPDATE,
            status_url="https://my-status-url.com",
        )

    # subcommand can also init in the same way
    with mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands:
        result = runner.invoke(
            "plus",
            "deploy",
            "start",
            "--yes",
            "--snapshot-base-condition",
            "on-update",
            "--status-url",
            "https://my-status-url.com",
        )
        assert result.exit_code == 0
        assert (
            "Deploying to the branch deployment for my-branch, with prod as the base deployment"
            in result.output
        )
        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(project.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            require_branch_deployment=True,
            git_url=None,
            location_name=[],
            dagster_env=None,
            snapshot_base_condition=SnapshotBaseDeploymentCondition.ON_UPDATE,
            status_url="https://my-status-url.com",
        )


def test_plus_deploy_cant_determine_branch(logged_in_dg_cli_config, project, runner, mocker):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value=None,
    )
    with mock_external_dagster_cloud_cli_command():
        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert result.exit_code == 0
        assert "Could not determine a git branch, so deploying to prod." in result.output


def test_plus_deploy_main_branch(logged_in_dg_cli_config, project, runner, mocker):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )
    with mock_external_dagster_cloud_cli_command():
        result = runner.invoke("plus", "deploy", "--agent-type", "serverless", "--yes")
        assert result.exit_code == 0
        assert "Current branch is main, so deploying to prod." in result.output


def test_plus_deploy_hybrid_no_build_yaml(logged_in_dg_cli_config, project, runner, mocker):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )
    with mock_external_dagster_cloud_cli_command():
        result = runner.invoke("plus", "deploy", "--agent-type", "hybrid", "--yes")

        assert result.exit_code

        assert "No build registry found. Please specify a registry key" in result.output


def test_plus_deploy_hybrid_with_yaml_files(
    logged_in_dg_cli_config, project, runner, mocker, build_yaml_file, container_context_yaml_file
):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )
    with mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands:
        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke("plus", "deploy", "--agent-type", "hybrid", "--yes")
            assert not result.exit_code, result.output

            mocked_cloud_cli_commands.init.assert_called_with(
                statedir=DEFAULT_STATEDIR_PATH,
                project_dir=str(project.resolve()),
                deployment="prod",
                organization="hooli",
                clean_statedir=False,
                commit_hash=None,
                dagster_cloud_yaml_path=mock.ANY,
                git_url=None,
                require_branch_deployment=False,
                dagster_env=None,
                location_name=[],
                snapshot_base_condition=None,
                status_url=None,
            )
            mocked_cloud_cli_commands.deploy.assert_called_once_with(
                statedir=DEFAULT_STATEDIR_PATH,
                agent_heartbeat_timeout=mock.ANY,
                location_load_timeout=mock.ANY,
                location_name=[],
            )

            dagster_cloud_yaml_path = DEFAULT_STATEDIR_PATH / Path("dagster_cloud.yaml")

            with open(dagster_cloud_yaml_path) as f:
                file = yaml.safe_load(f)
                assert file["locations"][0]["build"] == {
                    "directory": str(project.resolve()),
                    "registry": "my-repo",
                }
                assert file["locations"][0]["container_context"] == {
                    "k8s": {
                        "env_secrets": ["project-secret"],
                    },
                }


def test_plus_deploy_hybrid_with_workspace_yaml_files(
    logged_in_dg_cli_config,
    workspace,
    runner,
    mocker,
    workspace_build_yaml_file,
    workspace_container_context_yaml_file,
):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )

    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke("plus", "deploy", "--agent-type", "hybrid", "--yes")
            assert not result.exit_code, result.output

            dagster_cloud_yaml_path = DEFAULT_STATEDIR_PATH / Path("dagster_cloud.yaml")

            mocked_cloud_cli_commands.init.assert_called_with(
                statedir=DEFAULT_STATEDIR_PATH,
                project_dir=str(workspace.resolve()),
                deployment="prod",
                organization="hooli",
                clean_statedir=False,
                commit_hash=None,
                dagster_cloud_yaml_path=mock.ANY,
                git_url=None,
                require_branch_deployment=False,
                dagster_env=None,
                location_name=[],
                snapshot_base_condition=None,
                status_url=None,
            )

            mocked_cloud_cli_commands.deploy.assert_called_once_with(
                statedir=DEFAULT_STATEDIR_PATH,
                location_name=[],
                agent_heartbeat_timeout=mock.ANY,
                location_load_timeout=mock.ANY,
            )

            with open(dagster_cloud_yaml_path) as f:
                file = yaml.safe_load(f)
                assert file["locations"][0]["build"] == {
                    "directory": str(workspace.resolve()),  # from build.yaml
                    "registry": "my-workspace-repo",
                }
                assert file["locations"][0]["container_context"] == {
                    "k8s": {
                        "env_secrets": ["workspace-secret"],
                    },
                }


def test_plus_deploy_hybrid_with_merged_yaml_files(
    logged_in_dg_cli_config,
    workspace,
    runner,
    mocker,
    workspace_build_yaml_file,
    workspace_project_build_yaml_file,
    workspace_container_context_yaml_file,
    workspace_project_container_context_yaml_file,
):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )
    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke("plus", "deploy", "--agent-type", "hybrid", "--yes")
            assert not result.exit_code, result.output

            dagster_cloud_yaml_path = DEFAULT_STATEDIR_PATH / Path("dagster_cloud.yaml")

            mocked_cloud_cli_commands.init.assert_called_with(
                statedir=DEFAULT_STATEDIR_PATH,
                project_dir=str(workspace.resolve()),
                deployment="prod",
                organization="hooli",
                clean_statedir=False,
                commit_hash=None,
                dagster_cloud_yaml_path=mock.ANY,
                git_url=None,
                require_branch_deployment=False,
                dagster_env=None,
                location_name=[],
                snapshot_base_condition=None,
                status_url=None,
            )

            mocked_cloud_cli_commands.deploy.assert_called_once_with(
                statedir=DEFAULT_STATEDIR_PATH,
                location_name=[],
                agent_heartbeat_timeout=mock.ANY,
                location_load_timeout=mock.ANY,
            )

            with open(dagster_cloud_yaml_path) as f:
                file = yaml.safe_load(f)
                assert file["locations"][0]["build"] == {
                    "directory": str((workspace / "foo-bar").resolve()),  # from build.yaml
                    "registry": "my-project-repo",
                }

                assert file["locations"][0]["container_context"] == {
                    "k8s": {
                        "env_secrets": [
                            "workspace-secret",
                            "project-secret",
                        ],
                    },
                }


def test_plus_deploy_subcommands(
    logged_in_dg_cli_config, project, runner, mocker, build_yaml_file
) -> None:
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )

    with mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands:
        from dagster_cloud_cli.commands.ci import BuildStrategy

        result = runner.invoke("plus", "deploy", "start", "--yes")
        assert not result.exit_code, result.output
        assert "Current branch is main, so deploying to prod." in result.output

        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(project.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            git_url=None,
            require_branch_deployment=False,
            dagster_env=None,
            location_name=[],
            snapshot_base_condition=None,
            status_url=None,
        )

        mocked_cloud_cli_commands.reset_mocks()

        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke("plus", "deploy", "build-and-push", "--agent-type", "hybrid")
            assert not result.exit_code, result.output

        result = runner.invoke("plus", "deploy", "build-and-push", "--agent-type", "serverless")
        assert not result.exit_code, result.output
        mocked_cloud_cli_commands.build.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            build_directory=str(project.resolve()),
            dockerfile_path=str((project / "Dockerfile").absolute()),
            use_editable_dagster=False,
            build_strategy=BuildStrategy.docker,
            docker_image_tag=None,
            docker_base_image=None,
            docker_env=[],
            python_version=PYTHON_VERSION,
            pex_build_method=BuildMethod.DOCKER_FALLBACK,
            pex_deps_cache_from=None,
            pex_deps_cache_to=None,
            pex_base_image_tag=None,
            location_name=["foo-bar"],
        )

        mocked_cloud_cli_commands.reset_mocks()

        result = runner.invoke("plus", "deploy", "set-build-output", "--image-tag", "foo")
        assert not result.exit_code, result.output

        mocked_cloud_cli_commands.set_build_output.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            location_name=[],
            image_tag="foo",
        )

        mocked_cloud_cli_commands.reset_mocks()

        result = runner.invoke("plus", "deploy", "finish")
        assert not result.exit_code, result.output

        mocked_cloud_cli_commands.deploy.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            agent_heartbeat_timeout=mock.ANY,
            location_load_timeout=mock.ANY,
            location_name=[],
        )


def test_plus_deploy_subcommands_with_location(
    logged_in_dg_cli_config, project, runner, mocker, build_yaml_file
) -> None:
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )

    with mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands:
        from dagster_cloud_cli.commands.ci import BuildStrategy

        result = runner.invoke("plus", "deploy", "start", "--yes", "--location-name", "foo-bar")
        assert not result.exit_code, result.output
        assert "Current branch is main, so deploying to prod." in result.output

        mocked_cloud_cli_commands.init.assert_called_with(
            statedir=DEFAULT_STATEDIR_PATH,
            project_dir=str(project.resolve()),
            deployment="prod",
            organization="hooli",
            clean_statedir=False,
            dagster_cloud_yaml_path=mock.ANY,
            commit_hash=None,
            git_url=None,
            require_branch_deployment=False,
            dagster_env=None,
            location_name=["foo-bar"],
            snapshot_base_condition=None,
            status_url=None,
        )

        mocked_cloud_cli_commands.reset_mocks()

        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke(
                "plus",
                "deploy",
                "build-and-push",
                "--agent-type",
                "hybrid",
                "--location-name",
                "foo-bar",
            )
            assert not result.exit_code, result.output

        result = runner.invoke(
            "plus",
            "deploy",
            "build-and-push",
            "--agent-type",
            "serverless",
            "--location-name",
            "foo-bar",
        )
        mocked_cloud_cli_commands.build.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            build_directory=str(project.resolve()),
            dockerfile_path=str((project / "Dockerfile").absolute()),
            use_editable_dagster=False,
            build_strategy=BuildStrategy.docker,
            docker_image_tag=None,
            docker_base_image=None,
            docker_env=[],
            python_version=PYTHON_VERSION,
            pex_build_method=BuildMethod.DOCKER_FALLBACK,
            pex_deps_cache_from=None,
            pex_deps_cache_to=None,
            pex_base_image_tag=None,
            location_name=["foo-bar"],
        )

        mocked_cloud_cli_commands.reset_mocks()

        result = runner.invoke(
            "plus",
            "deploy",
            "set-build-output",
            "--image-tag",
            "foo",
            "--location-name",
            "foo-bar",
        )
        assert not result.exit_code, result.output
        mocked_cloud_cli_commands.set_build_output.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            location_name=["foo-bar"],
            image_tag="foo",
        )

        mocked_cloud_cli_commands.reset_mocks()

        result = runner.invoke("plus", "deploy", "finish", "--location-name", "foo-bar")
        assert not result.exit_code, result.output

        mocked_cloud_cli_commands.deploy.assert_called_once_with(
            statedir=DEFAULT_STATEDIR_PATH,
            agent_heartbeat_timeout=mock.ANY,
            location_load_timeout=mock.ANY,
            location_name=["foo-bar"],
        )


@responses.activate
def test_plus_deploy_hybrid_with_build_yaml_scaffold(
    logged_in_dg_cli_config, project, runner, mocker
):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )
    with mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands:
        mock_hybrid_response()

        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke("scaffold", "build-artifacts", "-y")
            assert not result.exit_code, result.output

            result = runner.invoke("plus", "deploy", "--agent-type", "hybrid", "--yes")
            assert not result.exit_code, result.output
            assert "No Dockerfile found" not in result.output

            mocked_cloud_cli_commands.init.assert_called_with(
                statedir=DEFAULT_STATEDIR_PATH,
                project_dir=str(project.resolve()),
                deployment="prod",
                organization="hooli",
                clean_statedir=False,
                commit_hash=None,
                dagster_cloud_yaml_path=mock.ANY,
                git_url=None,
                require_branch_deployment=False,
                dagster_env=None,
                location_name=[],
                snapshot_base_condition=None,
                status_url=None,
            )
            mocked_cloud_cli_commands.deploy.assert_called_once_with(
                statedir=DEFAULT_STATEDIR_PATH,
                agent_heartbeat_timeout=mock.ANY,
                location_load_timeout=mock.ANY,
                location_name=[],
            )


@responses.activate
def test_plus_deploy_hybrid_with_workspace_build_yaml_scaffold(
    logged_in_dg_cli_config,
    workspace,
    runner,
    mocker,
):
    mocker.patch(
        "dagster_dg_cli.cli.plus.deploy.deploy_session.get_local_branch_name",
        return_value="main",
    )

    mock_hybrid_response()

    with (
        mock_external_dagster_cloud_cli_command() as mocked_cloud_cli_commands,
    ):
        with patch(
            "dagster_dg_cli.cli.plus.deploy.deploy_session._build_hybrid_image",
        ):
            result = runner.invoke(
                "scaffold",
                "build-artifacts",
                "-y",
            )
            assert not result.exit_code, result.output

            result = runner.invoke("plus", "deploy", "--agent-type", "hybrid", "--yes")
            assert not result.exit_code, result.output
            assert "No Dockerfile found" not in result.output

            dagster_cloud_yaml_path = DEFAULT_STATEDIR_PATH / Path("dagster_cloud.yaml")

            mocked_cloud_cli_commands.init.assert_called_with(
                statedir=DEFAULT_STATEDIR_PATH,
                project_dir=str(workspace.resolve()),
                deployment="prod",
                organization="hooli",
                clean_statedir=False,
                commit_hash=None,
                dagster_cloud_yaml_path=mock.ANY,
                git_url=None,
                require_branch_deployment=False,
                dagster_env=None,
                location_name=[],
                snapshot_base_condition=None,
                status_url=None,
            )

            mocked_cloud_cli_commands.deploy.assert_called_once_with(
                statedir=DEFAULT_STATEDIR_PATH,
                location_name=[],
                agent_heartbeat_timeout=mock.ANY,
                location_load_timeout=mock.ANY,
            )

            with open(dagster_cloud_yaml_path) as f:
                assert yaml.safe_load(f)["locations"][0]["build"] == {
                    "directory": str(workspace.resolve() / "foo-bar"),  # from build.yaml
                    "registry": "...",
                }



"""Run API commands following GitHub CLI patterns."""

import json

import click
from dagster_dg_core.utils import DgClickCommand, DgClickGroup
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper
from dagster_shared.plus.config import DagsterPlusCliConfig
from dagster_shared.plus.config_utils import dg_api_options

from dagster_dg_cli.cli.api.client import create_dg_api_graphql_client


def format_run_table(run) -> str:
    """Format run as human-readable table."""
    lines = [
        f"Run ID: {run.id}",
        f"Status: {run.status.value}",
        f"Created: {run.created_at}",
    ]

    if run.started_at:
        lines.append(f"Started: {run.started_at}")
    if run.ended_at:
        lines.append(f"Ended: {run.ended_at}")
    if run.job_name:
        lines.append(f"Pipeline: {run.job_name}")

    return "\n".join(lines)


@click.command(name="get", cls=DgClickCommand)
@click.argument("run_id", type=str)
@click.option(
    "--json",
    "output_json",
    is_flag=True,
    help="Output in JSON format for machine readability",
)
@dg_api_options(deployment_scoped=True)
@cli_telemetry_wrapper
@click.pass_context
def get_run_command(
    ctx: click.Context,
    run_id: str,
    output_json: bool,
    organization: str,
    deployment: str,
    api_token: str,
    view_graphql: bool,
) -> None:
    """Get run metadata by ID."""
    from dagster_dg_cli.api_layer.api.run import DgApiRunApi

    config = DagsterPlusCliConfig.create_for_deployment(
        deployment=deployment,
        organization=organization,
        user_token=api_token,
    )
    client = create_dg_api_graphql_client(ctx, config, view_graphql=view_graphql)
    api = DgApiRunApi(client)

    try:
        run = api.get_run(run_id)

        if output_json:
            output = run.model_dump_json(indent=2)
        else:
            output = format_run_table(run)

        click.echo(output)
    except Exception as e:
        if output_json:
            error_response = {"error": str(e)}
            click.echo(json.dumps(error_response), err=True)
        else:
            click.echo(f"Error querying Dagster Plus API: {e}", err=True)


@click.group(
    name="run",
    cls=DgClickGroup,
    commands={
        "get": get_run_command,
    },
)
def run_group():
    """Manage runs in Dagster Plus."""



import json
import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from dagster_dg_core.utils import pushd
from dagster_shared.telemetry import (
    cleanup_telemetry_logger,
    get_or_create_dir_from_dagster_home,
    get_telemetry_logger,
)
from dagster_test.components.test_utils.test_cases import BASIC_INVALID_VALUE, BASIC_VALID_VALUE
from dagster_test.dg_utils.utils import (
    ProxyRunner,
    crawl_cli_commands,
    create_project_from_components,
    isolated_example_project_foo_bar,
    modify_environment_variable,
)

NO_TELEMETRY_COMMANDS = {
    ("utils", "inspect-component"),
    # Is actually instrumented, but since subcommands are dynamically generated we test manually
    ("scaffold",),
}


@pytest.fixture
def telemetry_caplog(caplog):
    # telemetry logger doesn't propagate to the root logger, so need to attach the caplog handler
    get_telemetry_logger().addHandler(caplog.handler)
    yield caplog
    get_telemetry_logger().removeHandler(caplog.handler)
    # Needed to avoid file contention issues on windows with the telemetry log file
    cleanup_telemetry_logger()


def test_telemetry_commands_properly_wrapped():
    commands = crawl_cli_commands()
    for command, command_defn in commands.items():
        if tuple(command[1:]) in NO_TELEMETRY_COMMANDS:
            continue

        fn = command_defn.callback
        while hasattr(fn, "__wrapped__"):
            if getattr(fn, "__has_cli_telemetry_wrapper", None) is True:
                break
            fn = getattr(fn, "__wrapped__")

        assert hasattr(fn, "__has_cli_telemetry_wrapper") is True, (
            f"Command {command} is not properly wrapped. Please wrap in the @cli_telemetry_wrapper decorator "
            "or add it to the NO_TELEMETRY_COMMANDS set."
        )


@pytest.mark.parametrize("success", [True, False])
def test_basic_logging_success_failure(
    telemetry_caplog: pytest.LogCaptureFixture, success: bool
) -> None:
    test_case = BASIC_VALID_VALUE if success else BASIC_INVALID_VALUE
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            test_case.component_path,
            local_component_defn_to_inject=test_case.component_type_filepath,
        ) as tmpdir,
        TemporaryDirectory() as dagster_home,
        modify_environment_variable("DAGSTER_HOME", dagster_home),
    ):
        telemetry_caplog.clear()
        with pushd(tmpdir):
            result = runner.invoke("check", "yaml")
            assert result.exit_code == 0 if success else 1

        assert os.path.exists(
            os.path.join(get_or_create_dir_from_dagster_home("logs"), "event.log")
        )
        assert len(telemetry_caplog.records) == 2, telemetry_caplog.records

        first_message = json.loads(telemetry_caplog.records[0].getMessage())
        second_message = json.loads(telemetry_caplog.records[1].getMessage())

        assert first_message["action"] == "check_yaml_command_started"
        assert second_message["action"] == "check_yaml_command_ended"
        assert second_message["metadata"]["command_success"] == "True" if success else "False", (
            second_message["metadata"]
        )


def test_telemetry_disabled_dagster_yaml(telemetry_caplog: pytest.LogCaptureFixture) -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_VALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_VALID_VALUE.component_type_filepath,
        ) as tmpdir,
        TemporaryDirectory() as dagster_home,
        modify_environment_variable("DAGSTER_HOME", dagster_home),
    ):
        telemetry_caplog.clear()

        dagster_yaml = Path(dagster_home) / "dagster.yaml"
        dagster_yaml.write_text(
            """
            telemetry:
                enabled: false
            """
        )
        with pushd(tmpdir):
            result = runner.invoke("check", "yaml")
            assert result.exit_code == 0

        assert not os.path.exists(
            os.path.join(get_or_create_dir_from_dagster_home("logs"), "event.log")
        )
        assert len(telemetry_caplog.records) == 0


def test_telemetry_disabled_dg_config(telemetry_caplog: pytest.LogCaptureFixture) -> None:
    with (
        ProxyRunner.test() as runner,
        create_project_from_components(
            runner,
            BASIC_VALID_VALUE.component_path,
            local_component_defn_to_inject=BASIC_VALID_VALUE.component_type_filepath,
        ) as tmpdir,
        TemporaryDirectory() as dg_cli_config_folder,
        TemporaryDirectory() as dagster_cloud_config_folder,
        modify_environment_variable("DG_CLI_CONFIG", str(Path(dg_cli_config_folder) / "dg.toml")),
        modify_environment_variable(
            "DAGSTER_CLOUD_CLI_CONFIG", str(Path(dagster_cloud_config_folder) / "config.yaml")
        ),
    ):
        telemetry_caplog.clear()
        dg_config_path = Path(dg_cli_config_folder) / "dg.toml"
        dg_config_path.write_text(
            """
            [cli.telemetry]
            enabled = false
            """
        )

        with pushd(tmpdir):
            result = runner.invoke("check", "yaml")
            assert result.exit_code == 0, str(result.exception)

        assert len(telemetry_caplog.records) == 0


@pytest.mark.skip("temp")
def test_telemetry_scaffold_component(telemetry_caplog: pytest.LogCaptureFixture) -> None:
    with (
        ProxyRunner.test(use_fixed_test_components=True) as runner,
        isolated_example_project_foo_bar(runner),
        TemporaryDirectory() as dagster_home,
        modify_environment_variable("DAGSTER_HOME", dagster_home),
    ):
        telemetry_caplog.clear()
        result = runner.invoke(
            "scaffold", "defs", "dagster_test.components.AllMetadataEmptyComponent", "qux"
        )
        assert result.exit_code == 0, result.output + " " + str(result.exception)
        assert Path("foo_bar/defs/qux").exists()
        assert len(telemetry_caplog.records) == 2
        first_message = json.loads(telemetry_caplog.records[0].getMessage())
        second_message = json.loads(telemetry_caplog.records[1].getMessage())
        assert first_message["action"] == "scaffold_component_command_started"
        assert second_message["action"] == "scaffold_component_command_ended"
        assert second_message["metadata"]["command_success"] == "True"



"""Schedule metadata schema definitions."""

from enum import Enum

from pydantic import BaseModel


class DgApiScheduleStatus(str, Enum):
    """Schedule execution status."""

    RUNNING = "RUNNING"
    STOPPED = "STOPPED"


class DgApiSchedule(BaseModel):
    """Single schedule metadata model."""

    id: str
    name: str
    status: DgApiScheduleStatus
    cron_schedule: str
    pipeline_name: str
    description: str | None = None
    execution_timezone: str | None = None
    code_location_origin: str | None = None
    next_tick_timestamp: float | None = None  # Unix timestamp

    class Config:
        from_attributes = True


class DgApiScheduleList(BaseModel):
    """GET /api/schedules response."""

    items: list[DgApiSchedule]
    total: int



import pytest


@pytest.fixture(autouse=True)
def clear_defined_commands():
    """Reset the _commands_defined flag on scaffold_group between tests,
    to ensure the cache of scaffold subcommands is cleared. This isn't an issue outside
    of tests because we're not reusing a Python process between different dg venvs.
    """
    from dagster_dg_cli.cli.scaffold import scaffold_defs_group

    scaffold_defs_group._commands_defined = False  # noqa: SLF001



from pathlib import Path

import pytest
import yaml

# Import the YAML generation functions from CLI
from dagster_dg_cli.utils.yaml_template_generator import (
    generate_defs_yaml_example_values,
    generate_defs_yaml_schema,
)

# Import the component classes directly
from dagster_test.components import SimplePipesScriptComponent

# Path to test data files for validation
TEST_DATA_DIR = Path(__file__).parent.parent / "yaml_template" / "test_data"

# Component test cases
TEST_COMPONENTS = [
    {
        "name": "simple_example",
        "component_class": SimplePipesScriptComponent,
        "test_data_dir": "simple_example",
    },
    # Note: Other components require additional dependencies that may not be available in test environment
    # They can be added when those dependencies are confirmed to be available
]


def _get_component_schema(component_class: type) -> dict:
    """Get the JSON schema for a component class."""
    model_cls = component_class.get_model_cls()
    if model_cls is None:
        return {}
    return model_cls.model_json_schema()


def _get_component_type_str(component_class: type) -> str:
    """Get the component type string for a component class."""
    return f"{component_class.__module__}.{component_class.__name__}"


@pytest.mark.parametrize("test_case", TEST_COMPONENTS)
def test_defs_yaml_schema_matches_test_data(test_case):
    """Test that generated YAML schema matches expected test data."""
    component_class = test_case["component_class"]
    test_data_path = TEST_DATA_DIR / test_case["test_data_dir"] / "expected_schema"

    # Skip if test data doesn't exist
    if not test_data_path.exists():
        pytest.skip(f"No expected_schema file for {test_case['name']}")

    # Generate YAML schema using the same function the CLI uses
    component_type_str = _get_component_type_str(component_class)
    component_schema = _get_component_schema(component_class)
    generated_yaml = generate_defs_yaml_schema(component_type_str, component_schema)

    # Load expected content
    expected_yaml = test_data_path.read_text().strip()

    assert generated_yaml.strip() == expected_yaml


@pytest.mark.parametrize("test_case", TEST_COMPONENTS)
def test_defs_yaml_example_values_matches_test_data(test_case):
    """Test that generated YAML example values match expected test data."""
    component_class = test_case["component_class"]
    test_data_path = TEST_DATA_DIR / test_case["test_data_dir"] / "expected_example"

    # Skip if test data doesn't exist
    if not test_data_path.exists():
        pytest.skip(f"No expected_example file for {test_case['name']}")

    # Generate YAML example values using the same function the CLI uses
    component_type_str = _get_component_type_str(component_class)
    component_schema = _get_component_schema(component_class)
    generated_yaml = generate_defs_yaml_example_values(component_type_str, component_schema)

    # Load expected content
    expected_yaml = test_data_path.read_text().strip()

    assert generated_yaml.strip() == expected_yaml


@pytest.mark.parametrize("test_case", TEST_COMPONENTS)
def test_generated_yaml_is_valid(test_case):
    """Test that both generated schema and example YAML are syntactically valid."""
    component_class = test_case["component_class"]
    component_type_str = _get_component_type_str(component_class)
    component_schema = _get_component_schema(component_class)

    # Test schema YAML (filter out comments)
    schema_yaml = generate_defs_yaml_schema(component_type_str, component_schema)
    schema_lines = [line for line in schema_yaml.split("\n") if not line.strip().startswith("#")]
    schema_content = "\n".join(schema_lines)

    try:
        yaml.safe_load(schema_content)
    except yaml.YAMLError as e:
        pytest.fail(f"Generated schema YAML is invalid for {test_case['name']}: {e}")

    # Test example YAML
    example_yaml = generate_defs_yaml_example_values(component_type_str, component_schema)
    try:
        parsed_example = yaml.safe_load(example_yaml)
        assert "type" in parsed_example
        assert "attributes" in parsed_example
    except yaml.YAMLError as e:
        pytest.fail(f"Generated example YAML is invalid for {test_case['name']}: {e}")


def test_component_type_strings_are_correct():
    """Test that component type strings match the expected format."""
    for test_case in TEST_COMPONENTS:
        component_class = test_case["component_class"]
        component_type_str = _get_component_type_str(component_class)
        component_schema = _get_component_schema(component_class)

        generated_schema = generate_defs_yaml_schema(component_type_str, component_schema)
        generated_example = generate_defs_yaml_example_values(component_type_str, component_schema)

        # Extract component type from generated YAML
        expected_type = f"{component_class.__module__}.{component_class.__name__}"

        # Check schema contains correct type
        assert f"type: {expected_type}" in generated_schema

        # Check example contains correct type
        assert f'type: "{expected_type}"' in generated_example


@pytest.mark.parametrize("test_case", TEST_COMPONENTS)
def test_schema_has_required_structure(test_case):
    """Test that generated schema has the expected structure elements."""
    component_class = test_case["component_class"]
    component_type_str = _get_component_type_str(component_class)
    component_schema = _get_component_schema(component_class)
    generated_yaml = generate_defs_yaml_schema(component_type_str, component_schema)

    # Should contain template comment
    assert "# Template with instructions" in generated_yaml

    # Should contain type and attributes sections
    assert "type:" in generated_yaml
    assert "attributes:" in generated_yaml

    # Should contain field type hints
    assert "<" in generated_yaml and ">" in generated_yaml

    # Should contain Required/Optional annotations
    assert "Required" in generated_yaml or "Optional" in generated_yaml


def test_all_test_data_files_correspond_to_components():
    """Ensure all test data directories have corresponding component definitions."""
    test_data_dirs = {d.name for d in TEST_DATA_DIR.iterdir() if d.is_dir()}
    component_test_dirs = {tc["test_data_dir"] for tc in TEST_COMPONENTS}

    # Note: It's okay to have more test data than components we test, since some
    # components may require dependencies not available in the test environment
    missing_test_data = component_test_dirs - test_data_dirs
    if missing_test_data:
        pytest.fail(f"Component definitions without test data directories: {missing_test_data}")



"""GraphQL implementation for agent operations."""

from typing import TYPE_CHECKING, Any, Optional

from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.agent import DgApiAgent, DgApiAgentList

# GraphQL queries
LIST_AGENTS_QUERY = """
query ListAgents {
    agents {
        id
        agentLabel
        status
        lastHeartbeatTime
        metadata {
            key
            value
        }
    }
}
"""

# Note: There's no single 'agent' query, so we use 'agents' and filter client-side


def process_agents_response(graphql_response: dict[str, Any]) -> "DgApiAgentList":
    """Process GraphQL response into AgentList.

    This is a pure function that can be easily tested without mocking GraphQL clients.

    Args:
        graphql_response: Raw GraphQL response containing "agents"

    Returns:
        AgentList: Processed agent data
    """
    # Import pydantic models only when needed
    from dagster_dg_cli.api_layer.schemas.agent import (
        DgApiAgent,
        DgApiAgentList,
        DgApiAgentMetadataEntry,
        DgApiAgentStatus,
    )

    agents_data = graphql_response.get("agents", [])

    agents = []
    for a in agents_data:
        metadata = [
            DgApiAgentMetadataEntry(key=m["key"], value=m["value"]) for m in a.get("metadata", [])
        ]

        agents.append(
            DgApiAgent(
                id=a["id"],
                agent_label=a.get("agentLabel"),
                status=DgApiAgentStatus[a["status"]],
                last_heartbeat_time=a.get("lastHeartbeatTime"),
                metadata=metadata,
            )
        )

    return DgApiAgentList(
        items=agents,
        total=len(agents),
    )


def list_agents_via_graphql(
    client: IGraphQLClient,
    limit: int | None = None,
) -> "DgApiAgentList":
    """Fetch agents using GraphQL.
    This is an implementation detail that can be replaced with REST calls later.
    """
    result = client.execute(LIST_AGENTS_QUERY)

    agent_list = process_agents_response(result)

    # Apply limit if specified
    if limit is not None:
        agent_list.items = agent_list.items[:limit]
        agent_list.total = len(agent_list.items)

    return agent_list


def get_agent_via_graphql(
    client: IGraphQLClient,
    agent_id: str,
) -> Optional["DgApiAgent"]:
    """Fetch single agent using GraphQL.
    This is an implementation detail that can be replaced with REST calls later.

    Note: Since there's no single 'agent' query, we fetch all agents and filter client-side.
    """
    result = client.execute(LIST_AGENTS_QUERY)
    agent_list = process_agents_response(result)

    # Find the agent with matching ID
    for agent in agent_list.items:
        if agent.id == agent_id:
            return agent

    return None



import os
import subprocess
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

import click
from dagster_cloud_cli.types import SnapshotBaseDeploymentCondition
from dagster_cloud_cli.utils import SUPPORTED_PYTHON_VERSIONS
from dagster_dg_core.config import DgRawCliConfig, normalize_cli_config
from dagster_dg_core.context import DgContext
from dagster_dg_core.shared_options import (
    dg_editable_dagster_options,
    dg_global_options,
    dg_path_options,
    make_option_group,
)
from dagster_dg_core.utils import DgClickCommand, DgClickGroup, activate_venv, not_none
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper
from dagster_shared import check
from dagster_shared.plus.config import DagsterPlusCliConfig
from dagster_shared.serdes import serialize_value
from dagster_shared.seven.temp_dir import get_system_temp_directory

from dagster_dg_cli.cli.plus.constants import DgPlusAgentType, DgPlusDeploymentType
from dagster_dg_cli.cli.plus.deploy.configure.commands import deploy_configure_group
from dagster_dg_cli.cli.plus.deploy.deploy_session import (
    build_artifact,
    finish_deploy_session,
    init_deploy_session,
)
from dagster_dg_cli.cli.plus.deploy.validation import _extract_dagster_env_from_url
from dagster_dg_cli.utils.plus.build import get_agent_type

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance
    from dagster_shared.serdes.objects.models.defs_state_info import DefsStateManagementType

DEFAULT_STATEDIR_PATH = os.path.join(get_system_temp_directory(), "dg-build-state")


def _get_statedir():
    return os.getenv("DAGSTER_BUILD_STATEDIR", DEFAULT_STATEDIR_PATH)


def _get_snapshot_base_deployment_conditions():
    # Lazy import to avoid loading dagster_cloud_cli at module import time

    return SnapshotBaseDeploymentCondition


def _get_organization(input_organization: str | None, plus_config: DagsterPlusCliConfig) -> str:
    organization = input_organization or plus_config.organization
    if not organization:
        raise click.UsageError(
            "Organization not specified. To specify an organization, use the --organization option "
            "or run `dg plus login`."
        )
    return organization


def _get_deployment(input_deployment: str | None, plus_config: DagsterPlusCliConfig) -> str:
    deployment = input_deployment or plus_config.default_deployment
    if not deployment:
        raise click.UsageError(
            "Deployment not specified. To specify a deployment, use the --deployment option "
            "or run `dg plus login`."
        )
    return deployment


org_and_deploy_option_group = make_option_group(
    {
        not_none(option.name): option
        for option in [
            click.Option(
                ["--organization"],
                "organization",
                help="Dagster+ organization to which to deploy. If not set, defaults to the value set by `dg plus login`.",
                envvar="DAGSTER_CLOUD_ORGANIZATION",
            ),
            click.Option(
                ["--deployment"],
                "deployment",
                help="Name of the Dagster+ deployment to which to deploy (or use as the base deployment if deploying to a branch deployment). If not set, defaults to the value set by `dg plus login`.",
                envvar="DAGSTER_CLOUD_DEPLOYMENT",
            ),
        ]
    }
)


build_strategy_option_group = make_option_group(
    {
        not_none(option.name): option
        for option in [
            click.Option(
                ["--build-strategy"],
                type=click.Choice(["docker", "python-executable"]),
                default="docker",
                help=(
                    "Build strategy used to build code locations. 'docker' builds a Docker image "
                    "(required for Hybrid agents). 'python-executable' builds PEX files "
                    "(Serverless agents only)."
                ),
                envvar="DAGSTER_BUILD_STRATEGY",
            ),
        ]
    }
)


pex_build_method_option_group = make_option_group(
    {
        not_none(option.name): option
        for option in [
            click.Option(
                ["--pex-build-method"],
                type=click.Choice(["local", "docker", "docker-fallback"]),
                default="docker-fallback",
                help=(
                    "Build method for PEX dependencies. 'docker-fallback' tries local first then Docker (default), "
                    "'local' uses only the current environment, 'docker' uses only a Docker builder. "
                    "Only applies when --build-strategy=python-executable."
                ),
                envvar="DAGSTER_PEX_BUILD_METHOD",
            ),
        ]
    }
)


@click.group(name="deploy", cls=DgClickGroup, invoke_without_command=True)
@org_and_deploy_option_group
@build_strategy_option_group
@pex_build_method_option_group
@click.option(
    "--python-version",
    "python_version",
    type=click.Choice(SUPPORTED_PYTHON_VERSIONS),
    help=(
        "Python version used to deploy the project. If not set, defaults to the calling process's Python minor version."
    ),
)
@click.option(
    "--deployment-type",
    "deployment_type_str",
    type=click.Choice([deployment_type.value for deployment_type in DgPlusDeploymentType]),
    help="Whether to deploy to a full deployment or a branch deployment. If unset, will attempt to infer from the current git branch.",
)
@click.option(
    "--agent-type",
    "agent_type_str",
    type=click.Choice([agent_type.value.lower() for agent_type in DgPlusAgentType]),
    help="Whether this a Hybrid or serverless code location.",
    required=False,
)
@click.option(
    "-y",
    "--yes",
    "skip_confirmation_prompt",
    is_flag=True,
    help="Skip confirmation prompts.",
)
@click.option("--git-url", "git_url")
@click.option("--commit-hash", "commit_hash")
@click.option(
    "--location-name",
    "location_names",
    help="Name of the code location to set the build output for. Defaults to the current project's code location, or every project's code location when run in a workspace.",
    required=False,
    multiple=True,
)
@click.option("--status-url", "status_url")
@click.option(
    "--snapshot-base-condition",
    "snapshot_base_condition_str",
    type=click.Choice(
        [
            snapshot_base_condition.value
            for snapshot_base_condition in _get_snapshot_base_deployment_conditions()
        ]
    ),
)
@click.option(
    "--skip-validation",
    is_flag=True,
    help="Skip configuration validation checks (not recommended).",
)
@dg_editable_dagster_options
@dg_path_options
@dg_global_options
@cli_telemetry_wrapper
def deploy_group(
    organization: str | None,
    deployment: str | None,
    build_strategy: str,
    pex_build_method: str,
    python_version: str | None,
    agent_type_str: str,
    deployment_type_str: str | None,
    git_url: str | None,
    commit_hash: str | None,
    use_editable_dagster: str | None,
    skip_confirmation_prompt: bool,
    location_names: tuple[str],
    target_path: Path,
    status_url: str | None,
    snapshot_base_condition_str: str | None,
    skip_validation: bool,
    **global_options: object,
) -> None:
    """Deploy a project or workspace to Dagster Plus. Handles all state management for the deploy
    session, building and pushing a new code artifact for each project.

    To run a full end-to-end deploy, run `dg plus deploy`. This will start a new session, build
    and push the image for the project or workspace, and inform Dagster+ to deploy the newly built
    code.

    Each of the individual stages of the deploy is also available as its own subcommand for additional
    customization.
    """
    from dagster_cloud_cli.commands.ci import BuildStrategy
    from dagster_cloud_cli.core.pex_builder.deps import BuildMethod

    if click.get_current_context().invoked_subcommand:
        return

    snapshot_base_condition = (
        _get_snapshot_base_deployment_conditions()(snapshot_base_condition_str)
        if snapshot_base_condition_str
        else None
    )

    cli_config = normalize_cli_config(global_options, click.get_current_context())
    plus_config = (
        DagsterPlusCliConfig.get() if DagsterPlusCliConfig.exists() else DagsterPlusCliConfig()
    )
    organization = _get_organization(organization, plus_config)
    deployment = _get_deployment(deployment, plus_config)

    # Extract dagster_env from config
    dagster_env = None
    if DagsterPlusCliConfig.exists():
        config = DagsterPlusCliConfig.get()
        dagster_env = _extract_dagster_env_from_url(config.url)

    dg_context = DgContext.for_workspace_or_project_environment(target_path, cli_config)
    _validate_location_names(dg_context, location_names, cli_config)

    # TODO Confirm that dagster-cloud is packaged in the project

    statedir = _get_statedir()

    if agent_type_str:
        agent_type = DgPlusAgentType(agent_type_str.upper())
    else:
        agent_type = get_agent_type(plus_config)

    build_strategy_enum = BuildStrategy(build_strategy)
    pex_build_method_enum = BuildMethod(pex_build_method)

    init_deploy_session(
        organization,
        deployment,
        dg_context,
        statedir,
        DgPlusDeploymentType(deployment_type_str) if deployment_type_str else None,
        skip_confirmation_prompt,
        git_url,
        commit_hash,
        location_names,
        status_url,
        snapshot_base_condition,
        dagster_env,
        skip_validation,
    )

    build_artifact(
        dg_context,
        agent_type,
        build_strategy_enum,
        pex_build_method_enum,
        statedir,
        bool(use_editable_dagster),
        python_version,
        location_names,
    )

    finish_deploy_session(dg_context, statedir, location_names)


def _validate_location_names(
    dg_context: DgContext, location_names: tuple[str], cli_config: DgRawCliConfig
):
    if not location_names:
        return

    if dg_context.is_project:
        existing_location_names = {dg_context.code_location_name}
    else:
        existing_location_names = {
            dg_context.for_project_environment(project_spec.path, cli_config).code_location_name
            for project_spec in dg_context.project_specs
        }
    nonexistent_location_names = set(location_names) - existing_location_names
    if nonexistent_location_names:
        raise click.UsageError(
            f"The following requested locations do not exist: {', '.join(nonexistent_location_names)}"
        )


@deploy_group.command(name="start", cls=DgClickCommand)
@org_and_deploy_option_group
@click.option(
    "--deployment-type",
    "deployment_type_str",
    type=click.Choice([deployment_type.value for deployment_type in DgPlusDeploymentType]),
    help="Whether to deploy to a full deployment or a branch deployment. If unset, will attempt to infer from the current git branch.",
)
@click.option(
    "-y",
    "--yes",
    "skip_confirmation_prompt",
    is_flag=True,
    help="Skip confirmation prompts.",
)
@click.option("--git-url", "git_url")
@click.option("--commit-hash", "commit_hash")
@click.option(
    "--location-name",
    "location_names",
    help="Name of the code location to set the build output for. Defaults to the current project's code location, or every project's code location when run in a workspace.",
    required=False,
    multiple=True,
)
@dg_path_options
@click.option("--status-url", "status_url")
@click.option(
    "--snapshot-base-condition",
    "snapshot_base_condition_str",
    type=click.Choice(
        [
            snapshot_base_condition.value
            for snapshot_base_condition in _get_snapshot_base_deployment_conditions()
        ]
    ),
)
@click.option(
    "--skip-validation",
    is_flag=True,
    help="Skip configuration validation checks (not recommended).",
)
@dg_global_options
@cli_telemetry_wrapper
def start_deploy_session_command(
    organization: str | None,
    deployment: str | None,
    deployment_type_str: str | None,
    skip_confirmation_prompt: bool,
    git_url: str | None,
    commit_hash: str | None,
    location_names: tuple[str],
    target_path: Path,
    status_url: str | None,
    snapshot_base_condition_str: str | None,
    skip_validation: bool,
    **global_options: object,
) -> None:
    """Start a new deploy session. Determines which code locations will be deployed and what
    deployment is being targeted (creating a new branch deployment if needed), and initializes a
    folder on the filesystem where state about the deploy session will be stored.
    """
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    plus_config = (
        DagsterPlusCliConfig.get() if DagsterPlusCliConfig.exists() else DagsterPlusCliConfig()
    )
    organization = _get_organization(organization, plus_config)
    deployment = _get_deployment(deployment, plus_config)

    # Extract dagster_env from config
    dagster_env = None
    if DagsterPlusCliConfig.exists():
        config = DagsterPlusCliConfig.get()
        dagster_env = _extract_dagster_env_from_url(config.url)

    dg_context = DgContext.for_workspace_or_project_environment(target_path, cli_config)
    _validate_location_names(dg_context, location_names, cli_config)
    statedir = _get_statedir()

    snapshot_base_condition = (
        _get_snapshot_base_deployment_conditions()(snapshot_base_condition_str)
        if snapshot_base_condition_str
        else None
    )

    init_deploy_session(
        organization,
        deployment,
        dg_context,
        statedir,
        DgPlusDeploymentType(deployment_type_str) if deployment_type_str else None,
        skip_confirmation_prompt,
        git_url,
        commit_hash,
        location_names,
        status_url,
        snapshot_base_condition,
        dagster_env,
        skip_validation,
    )


@deploy_group.command(name="build-and-push", cls=DgClickCommand)
@click.option(
    "--agent-type",
    "agent_type_str",
    type=click.Choice([agent_type.value.lower() for agent_type in DgPlusAgentType]),
    help="Whether this a Hybrid or serverless code location.",
)
@build_strategy_option_group
@pex_build_method_option_group
@click.option(
    "--python-version",
    "python_version",
    type=click.Choice(SUPPORTED_PYTHON_VERSIONS),
    help=(
        "Python version used to deploy the project. If not set, defaults to the calling process's Python minor version."
    ),
)
@click.option(
    "--location-name",
    "location_names",
    help="Name of the code location to set the build output for. Defaults to the current project's code location, or every project's code location when run in a workspace.",
    required=False,
    multiple=True,
)
@dg_editable_dagster_options
@dg_path_options
@dg_global_options
@cli_telemetry_wrapper
def build_and_push_command(
    agent_type_str: str,
    build_strategy: str,
    pex_build_method: str,
    python_version: str | None,
    use_editable_dagster: str | None,
    location_names: tuple[str],
    target_path: Path,
    **global_options: object,
) -> None:
    """Builds a Docker image to be deployed, and pushes it to the registry
    that was configured when the deploy session was started.
    """
    from dagster_cloud_cli.commands.ci import BuildStrategy
    from dagster_cloud_cli.core.pex_builder.deps import BuildMethod

    cli_config = normalize_cli_config(global_options, click.get_current_context())

    dg_context = DgContext.for_workspace_or_project_environment(target_path, cli_config)

    _validate_location_names(dg_context, location_names, cli_config)

    if agent_type_str:
        agent_type = DgPlusAgentType(agent_type_str.upper())
    else:
        plus_config = DagsterPlusCliConfig.get()
        agent_type = get_agent_type(plus_config)

    build_strategy_enum = BuildStrategy(build_strategy)
    pex_build_method_enum = BuildMethod(pex_build_method)

    statedir = _get_statedir()

    build_artifact(
        dg_context,
        agent_type,
        build_strategy_enum,
        pex_build_method_enum,
        statedir,
        bool(use_editable_dagster),
        python_version,
        location_names,
    )


@contextmanager
def _instance_with_defs_state_storage(
    ctx: click.Context, location_state
) -> Iterator["DagsterInstance"]:
    """Create a temporary instance with DagsterPlusCliDefsStateStorage configuration based off of the given LocationState."""
    import yaml
    from dagster._core.instance import DagsterInstance
    from dagster_cloud_cli.config_utils import get_organization, get_user_token

    api_token = check.not_none(get_user_token(ctx))
    organization = check.not_none(get_organization(ctx))

    # create dagster.yaml config
    dagster_config = {
        "defs_state_storage": {
            "module": "dagster_dg_cli.utils.plus.defs_state_storage",
            "class": "DagsterPlusCliDefsStateStorage",
            "config": {
                "url": location_state.url,
                "api_token": api_token,
                "deployment": location_state.deployment_name,
                "organization": organization,
            },
        }
    }

    with tempfile.TemporaryDirectory() as temp_dir:
        dagster_yaml_path = Path(temp_dir) / "dagster.yaml"
        dagster_yaml_path.write_text(yaml.dump(dagster_config))
        with DagsterInstance.from_config(temp_dir) as instance:
            yield instance


def refresh_defs_state_impl(
    ctx: click.Context,
    statedir: str,
    dg_context: DgContext,
    cli_config: DgRawCliConfig,
    management_types: set["DefsStateManagementType"],
):
    from dagster_cloud_cli.commands.ci import state

    state_store = state.FileStore(statedir=statedir)
    locations = state_store.list_selected_locations()

    if not locations:
        click.echo("No locations to refresh.")
        return

    # Determine which projects/locations to process
    if dg_context.is_project:
        # Single project - process it with each matching location_state
        project_contexts = [(dg_context, location_state) for location_state in locations]
    else:
        # Workspace - match projects to location states
        project_contexts = []
        for location_state in locations:
            # Find matching project by location_name
            for project_spec in dg_context.project_specs:
                project_context = dg_context.for_project_environment(project_spec.path, cli_config)
                if project_context.code_location_name == location_state.location_name:
                    project_contexts.append((project_context, location_state))
                    break

    # Create temp instance with defs_state_storage - shared across all locations
    # We'll use the first location_state to create the instance
    with _instance_with_defs_state_storage(ctx, locations[0]) as instance:
        instance_ref = instance.get_ref()
        serialized_instance_ref = serialize_value(instance_ref)

        # Run subprocess for each project/location
        for project_context, location_state in project_contexts:
            python_executable = project_context.project_python_executable

            # Build command
            cmd = [
                str(python_executable),
                "-m",
                "dagster_dg_cli.cli.entrypoint",
                "utils",
                "refresh-defs-state",
                "--instance-ref",
                serialized_instance_ref,
                "--target-path",
                str(project_context.root_path),
            ]

            # Add management-type args if specified
            for mt in management_types:
                cmd.extend(["--management-type", mt.value])

            click.echo(
                f"Refreshing defs state for location: {location_state.location_name} with command: {cmd}"
            )

            try:
                # activate the venv for the subprocess to ensure CLIs are available
                with activate_venv(project_context.root_path / ".venv"):
                    subprocess.run(cmd, check=True, capture_output=False)
            except subprocess.CalledProcessError as e:
                click.echo(
                    click.style(
                        f"Failed to refresh defs state for {location_state.location_name}: {e}",
                        fg="red",
                    )
                )
                raise

        # After all subprocesses complete, get the final DefsStateInfo
        defs_state_storage = check.not_none(instance.defs_state_storage)
        final_defs_state_info = defs_state_storage.get_latest_defs_state_info()

        if final_defs_state_info is None:
            click.echo("No defs_state_info to update")
            return

        # Set the same DefsStateInfo on all location states
        for location_state in locations:
            location_state.defs_state_info = final_defs_state_info
            state_store.save(location_state)

    click.echo(f"Updated defs_state_info for all {len(locations)} locations.")


@deploy_group.command(name="refresh-defs-state", cls=DgClickCommand)
@dg_editable_dagster_options
@dg_path_options
@dg_global_options
@click.option(
    "--management-type",
    multiple=True,
    type=click.Choice(["LOCAL_FILESYSTEM", "VERSIONED_STATE_STORAGE"]),
    help="Only refresh components with the specified management type. Can be specified multiple times to include multiple types. By default, refreshes VERSIONED_STATE_STORAGE and LOCAL_FILESYSTEM components.",
)
@cli_telemetry_wrapper
def refresh_defs_state_command(
    target_path: Path,
    management_type: tuple[str, ...],
    **global_options: object,
):
    """[Experimental] If using StateBackedComponents, this command will execute the `refresh_state` on each of them,
    and set the defs_state_info for each location.
    """
    from dagster_shared.serdes.objects.models.defs_state_info import DefsStateManagementType

    ctx = click.get_current_context()
    cli_config = normalize_cli_config(global_options, ctx)
    # Support both workspace and project contexts
    dg_context = DgContext.for_workspace_or_project_environment(target_path, cli_config)
    management_types = (
        {DefsStateManagementType(mt) for mt in management_type}
        if management_type
        else {
            DefsStateManagementType.VERSIONED_STATE_STORAGE,
            DefsStateManagementType.LOCAL_FILESYSTEM,
        }
    )
    refresh_defs_state_impl(
        ctx=ctx,
        statedir=_get_statedir(),
        dg_context=dg_context,
        cli_config=cli_config,
        management_types=management_types,
    )


@deploy_group.command(name="set-build-output", cls=DgClickCommand)
@click.option(
    "--image-tag",
    "image_tag",
    help="Tag for the built docker image.",
    required=True,
)
@click.option(
    "--location-name",
    "location_names",
    help="Name of the code location to set the build output for. Defaults to the current project's code location, or every project's code location when run in a workspace.",
    required=False,
    multiple=True,
)
@dg_global_options
@dg_path_options
@cli_telemetry_wrapper
def set_build_output_command(
    image_tag: str, location_names: tuple[str], target_path: Path, **global_options: object
) -> None:
    """If building a Docker image was built outside of the `dg` CLI, configures the deploy session
    to indicate the correct tag to use when the session is finished.
    """
    # Lazy import to avoid loading dagster_cloud_cli at module import time
    from dagster_cloud_cli.commands.ci import set_build_output

    cli_config = normalize_cli_config(global_options, click.get_current_context())

    dg_context = DgContext.for_workspace_or_project_environment(target_path, cli_config)
    statedir = _get_statedir()
    _validate_location_names(dg_context, location_names, cli_config)

    set_build_output(
        statedir=str(statedir),
        location_name=list(location_names),
        image_tag=image_tag,
    )


@deploy_group.command(name="finish", cls=DgClickCommand)
@click.option(
    "--location-name",
    "location_names",
    help="Name of the code location to set the build output for. Defaults to the current project's code location, or every project's code location when run in a workspace.",
    required=False,
    multiple=True,
)
@dg_global_options
@dg_path_options
@cli_telemetry_wrapper
def finish_deploy_session_command(
    location_names: tuple[str], target_path: Path, **global_options: object
) -> None:
    """Once all needed images have been built and pushed, completes the deploy session, signaling
    to the Dagster+ API that the deployment can be updated to the newly built and pushed code.
    """
    cli_config = normalize_cli_config(global_options, click.get_current_context())

    dg_context = DgContext.for_workspace_or_project_environment(target_path, cli_config)
    _validate_location_names(dg_context, location_names, cli_config)
    statedir = _get_statedir()

    finish_deploy_session(dg_context, statedir, location_names)


# Register the configure subcommand group
deploy_group.add_command(deploy_configure_group)



import json
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import click
from dagster_shared.plus.config import DagsterPlusCliConfig

if TYPE_CHECKING:
    from dagster_cloud_cli.commands.ci.state import LocationState


class IGraphQLClient(ABC):
    """Abstract base class for GraphQL clients with execute method."""

    @abstractmethod
    def execute(self, query: str, variables: Mapping[str, Any] | None = None) -> dict: ...


class DagsterPlusUnauthorizedError(click.ClickException):
    pass


class DebugGraphQLClient(IGraphQLClient):
    """GraphQL client wrapper that logs queries and responses to stderr for debugging."""

    def __init__(self, wrapped_client: IGraphQLClient):
        self.wrapped_client = wrapped_client

    def execute(self, query: str, variables: Mapping[str, Any] | None = None) -> dict:
        """Execute GraphQL query and log details to stderr."""
        # Print query details to stderr
        click.echo("=== GraphQL Query ===", err=True)
        click.echo(query, err=True)

        if variables:
            click.echo("=== Variables ===", err=True)
            click.echo(json.dumps(dict(variables), indent=2), err=True)

        # Execute the actual query
        result = self.wrapped_client.execute(query, variables)

        # Print response to stderr
        click.echo("=== Response ===", err=True)
        click.echo(json.dumps(result, indent=2), err=True)
        click.echo("=" * 20, err=True)

        return result


class DagsterPlusGraphQLClient(IGraphQLClient):
    def __init__(self, url: str, headers: Mapping[str, str]):
        # defer for import performance
        from gql import Client
        from gql.transport.requests import RequestsHTTPTransport

        self.client = Client(
            transport=RequestsHTTPTransport(url=url, use_json=True, headers=dict(headers))
        )

    @classmethod
    def from_config(cls, config: DagsterPlusCliConfig):
        return cls(
            url=f"{config.organization_url}/graphql",
            headers={
                k: v
                for k, v in {
                    "Dagster-Cloud-Api-Token": config.user_token,
                    "Dagster-Cloud-Organization": config.organization,
                    "Dagster-Cloud-Deployment": config.default_deployment,
                }.items()
                if v is not None
            },
        )

    @classmethod
    def from_location_state(
        cls, location_state: "LocationState", api_token: str, organization: str
    ):
        return cls(
            url=f"{location_state.url}/graphql",
            headers={
                "Dagster-Cloud-Api-Token": api_token,
                "Dagster-Cloud-Organization": organization,
                "Dagster-Cloud-Deployment": location_state.deployment_name,
            },
        )

    def execute(self, query: str, variables: Mapping[str, Any] | None = None):
        # defer for import performance
        from gql import gql

        result = self.client.execute(gql(query), variable_values=dict(variables or {}))
        value = next(iter(result.values()))
        if isinstance(value, Mapping):
            if value.get("__typename") == "UnauthorizedError":
                raise DagsterPlusUnauthorizedError("Unauthorized: " + value["message"])
            elif value.get("__typename", "").endswith("Error"):
                raise click.ClickException("Error: " + value["message"])
        return result



import tempfile
from pathlib import Path

import pytest


@pytest.fixture()
def dg_plus_cli_config(monkeypatch):
    with (
        tempfile.TemporaryDirectory() as tmp_dg_dir,
        tempfile.TemporaryDirectory() as tmp_cloud_dir,
    ):
        config_path = Path(tmp_dg_dir) / "dg.toml"
        config_path.write_text(
            """
            [cli.plus]
            organization = "hooli"
            user_token = "abc123"
            default_deployment = "hooli-dev"
            """
        )
        monkeypatch.setenv("DG_CLI_CONFIG", config_path)
        monkeypatch.setenv("DAGSTER_CLOUD_CLI_CONFIG", str(Path(tmp_cloud_dir) / "config"))
        yield config_path



import subprocess
from collections.abc import Sequence

from mcp.server.fastmcp import FastMCP

mcp = FastMCP("dagster-dg-cli")


def _subprocess(command: Sequence[str], cwd: str | None = None) -> str:
    """Execute a subprocess command and return decoded output.

    Wrapper around subprocess.check_output that captures both stdout and stderr,
    providing better error context for MCP tool calls by exposing the full command
    output when errors occur.

    Args:
        command: Sequence of command arguments to execute (e.g., ['dg', 'list', 'components']).
        cwd: Working directory path where the command should be executed.

    Returns:
        Decoded UTF-8 string output from the executed command.

    Raises:
        Exception: If the subprocess fails, raises an exception with the command's error output.
    """
    try:
        return subprocess.check_output(
            command,
            cwd=cwd,
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
    except subprocess.CalledProcessError as e:
        raise Exception(e.output)


@mcp.tool()
async def list_available_components(project_path: str) -> str:
    """List all available Dagster components that can be scaffolded in the project.

    This tool provides a comprehensive list of all component types available for scaffolding
    in the specified Dagster project. Components are reusable building blocks that can be
    used to create assets, jobs, schedules, and other Dagster definitions. Use this tool
    when you need to discover what components are available before scaffolding new definitions.

    The output includes component names, types, and brief descriptions to help you choose
    the appropriate component for your use case.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).

    Returns:
        JSON-formatted string containing the list of available components with their metadata,
        including component names, types, and descriptions.
    """
    return _subprocess(
        [
            "dg",
            "list",
            "components",
            "--json",
        ],
        cwd=project_path,
    )


@mcp.tool()
async def scaffold_dagster_component_help(
    project_path: str,
    component_type: str,
) -> str:
    """Get detailed help and parameter requirements for scaffolding a specific component type.

    This tool provides comprehensive documentation about the parameters, options, and usage
    patterns for a specific component type. Use this before scaffolding a component to
    understand what arguments are required and what configuration options are available.

    The help output includes parameter descriptions, default values, validation rules,
    and example usage patterns specific to the component type.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).
        component_type: Fully qualified component type identifier (e.g.,
                       'dagster_sling.SlingReplicationCollectionComponent',
                       'dagster_dbt.DbtProjectComponent'). Use list_available_components
                       to discover available component types.

    Returns:
        Detailed help text describing the component's parameters, usage patterns,
        and configuration options. This includes required and optional parameters,
        their types, default values, and example configurations.
    """
    return _subprocess(
        [
            "dg",
            "scaffold",
            "defs",
            component_type,
            "--help",
        ],
        cwd=project_path,
    )


@mcp.tool()
async def scaffold_dagster_component(
    project_path: str, component_type: str, component_name: str, component_arguments: list[str]
) -> str:
    """Scaffold a new Dagster component with the specified configuration.

    This tool creates a new component directory structure with a defs.yaml file that can be
    customized with component-specific configuration. The scaffolding process sets up the
    basic file structure and configuration template needed for the component to function.

    After scaffolding, you'll typically need to:
    1. Edit the generated defs.yaml file with your specific configuration
    2. Run check_dagster_defs_yaml to validate the YAML syntax
    3. Run check_dagster_definitions to ensure the component loads correctly

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).
        component_type: Fully qualified component type identifier (e.g.,
                       'dagster_sling.SlingReplicationCollectionComponent').
                       Use list_available_components to find available types.
        component_name: Name for the new component instance. This will be used as the
                       directory name and component identifier. Should be a valid
                       Python identifier (letters, numbers, underscores only).
        component_arguments: List of command-line arguments to pass to the scaffold command.
                           Use scaffold_dagster_component_help to determine what arguments
                           are available for the specific component type.
                           Example: ['--connection-name', 'my_db', '--table-pattern', 'users_*']

    Returns:
        Output from the scaffold command, including the path to the created component
        directory and any additional setup instructions or warnings.
    """
    return _subprocess(
        [
            "dg",
            "scaffold",
            "defs",
            component_type,
            component_name,
            *component_arguments,
        ],
        cwd=project_path,
    )


@mcp.tool()
async def inspect_component_type(project_path: str, component_type: str) -> str:
    """Inspect a component type to understand its YAML schema, configuration options, and metadata.

    This tool provides detailed information about a component type's structure, including:
    - The description of the component type.
    - The jsonschema that yaml files that use this type must abide by.

    Use this tool before editing a component's defs.yaml file to understand what
    configuration options are available and how to structure the YAML correctly.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).
        component_type: Fully qualified component type identifier to inspect (e.g.,
                       'dagster_sling.SlingReplicationCollectionComponent',
                       'dagster_dbt.DbtProjectComponent'). Use list_available_components
                       to discover available component types.

    Returns:
        Detailed inspection output including the component's YAML schema definition,
        configuration parameters with their types and constraints, and metadata about
        the component's behavior and requirements.
    """
    return _subprocess(
        [
            "dg",
            "utils",
            "inspect-component",
            component_type,
        ],
        cwd=project_path,
    )


@mcp.tool()
async def check_dagster_defs_yaml(project_path: str) -> str:
    """Validate the syntax and structure of all defs.yaml files in the project.

    This tool performs comprehensive validation of YAML syntax, schema compliance,
    and structural correctness for all component definition files in the project.
    It checks for:
    - Valid YAML syntax (proper indentation, quotes, etc.)
    - Schema compliance (required fields, correct data types)
    - Reference integrity (valid component type names)

    Run this tool after making any changes to defs.yaml files to catch syntax
    errors and configuration issues before attempting to load the definitions.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).

    Returns:
        Validation results indicating whether all defs.yaml files are syntactically
        correct and properly structured. If errors are found, detailed error messages
        with file paths and line numbers are provided to help with debugging.
    """
    return _subprocess(
        [
            "dg",
            "check",
            "yaml",
        ],
        cwd=project_path,
    )


@mcp.tool()
async def check_dagster_definitions(project_path: str) -> str:
    """Validate that all Dagster definitions can be successfully loaded and instantiated.

    This tool performs a comprehensive validation of the entire project's definitions by
    attempting to load all components and build their definitions.

    This is a deeper validation than check_dagster_defs_yaml - while that tool only
    checks YAML syntax and schema, this tool actually loads and runs the Python code and
    verifies that all components can be instantiated with the provided configuration.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).

    Returns:
        Validation results indicating whether all definitions can be successfully loaded.
        If loading fails, detailed error messages with component names, error types,
        and troubleshooting guidance are provided.
    """
    return _subprocess(
        [
            "dg",
            "check",
            "defs",
        ],
        cwd=project_path,
    )


@mcp.tool()
async def list_dagster_definitions(project_path: str) -> str:
    """Retrieve comprehensive metadata about all definitions in the Dagster project.

    This tool provides detailed information about all successfully loaded definitions
    in the project.

    Use this tool to get an overview of all definitions and where they were defined.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).

    Returns:
        JSON-formatted metadata about all definitions in the project, including
        their names, types, dependencies, and configuration details. The structure
        provides a comprehensive view of the project's data pipeline architecture.
    """
    return _subprocess(
        [
            "dg",
            "list",
            "defs",
            "--json",
        ],
        cwd=project_path,
    )


@mcp.tool()
async def list_available_integrations(project_path: str) -> str:
    """Retrieve an index of all available Dagster integrations in the marketplace.

    This provides information on integrations in the marketplace that are available for
    installation. It is a helpful resource for determine which tools are available, and provides
    context for determining which integration may be relevant when scaffolding components.

    Args:
        project_path: Absolute filesystem path to the root directory of your Dagster project
                     (the directory containing pyproject.toml with [tool.dg] or dg.toml).

    Returns:
        JSON-formatted metadata about all integrations available for installation, including
        their name, description, and pypi installation url. This resource is useful for determining
        which integrations to use when scaffolding pipelines.
    """
    return _subprocess(
        [
            "dg",
            "utils",
            "integrations",
            "--json",
        ],
        cwd=project_path,
    )


if __name__ == "__main__":
    mcp.run(transport="stdio")



"""Run metadata API implementation."""

from dataclasses import dataclass
from typing import TYPE_CHECKING

from dagster_dg_cli.api_layer.graphql_adapter.run import get_run_via_graphql
from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.run import DgApiRun


@dataclass(frozen=True)
class DgApiRunApi:
    """API for run metadata operations."""

    client: IGraphQLClient

    def get_run(self, run_id: str) -> "DgApiRun":
        """Get run metadata by ID."""
        return get_run_via_graphql(self.client, run_id)



import shutil
import subprocess

from dagster_dg_cli.cli import cli
from dagster_dg_core.utils import DgClickCommand, DgClickGroup
from dagster_test.dg_utils.utils import ProxyRunner, isolated_dg_venv


# Important that all nodes of the command tree inherit from one of our customized click
# Command/Group subclasses to ensure that the help formatting is consistent.
def test_all_commands_custom_subclass():
    def crawl(command):
        assert isinstance(command, (DgClickGroup, DgClickCommand)), (
            f"Group is not a DgClickGroup or DgClickCommand: {command}"
        )
        if isinstance(command, DgClickGroup):
            for command in command.commands.values():
                crawl(command)

    crawl(cli)


# Important that all nodes of the command tree inherit from one of our customized click
# Command/Group subclasses to ensure that the help formatting is consistent.
def test_all_commands_have_global_options():
    def crawl(command):
        assert isinstance(command, (DgClickGroup, DgClickCommand)), (
            f"Group is not a DgClickGroup or DgClickCommand: {command}"
        )
        if isinstance(command, DgClickGroup):
            for command in command.commands.values():
                crawl(command)

    crawl(cli)


# Install dagster-dg-cli in an isolated venv and make sure that it can execute. Without this test, it is
# easy to accidentally leave an import of a dependency present in the test environment inside
# dagster-dg-cli, which causes the executable to fail when it is installed from pypi without those test
# dependencies.
def test_isolated_dg_executes():
    with ProxyRunner.test() as runner, isolated_dg_venv(runner) as venv_path:
        dg_path = shutil.which("dg")
        assert dg_path is not None, "dg executable not found in PATH"
        assert dg_path.startswith(str(venv_path)), "dg executable not resolving to local venv"
        assert subprocess.check_output(["dg", "--help"])



"""Shared utilities for API commands."""

import click
from dagster_shared.plus.config import DagsterPlusCliConfig

from dagster_dg_cli.cli.api.client import DgApiTestContext


def get_config_or_error() -> DagsterPlusCliConfig:
    """Get Dagster Plus config or raise error if not authenticated."""
    if not DagsterPlusCliConfig.exists():
        raise click.UsageError(
            "`dg api` commands require authentication with Dagster Plus. Run `dg plus login` to authenticate."
        )
    return DagsterPlusCliConfig.get()


def get_config_for_api_command(ctx: click.Context) -> DagsterPlusCliConfig:
    """Get config for API commands, supporting both test and normal contexts."""
    # Check if we're in a test context
    if ctx.obj and isinstance(ctx.obj, DgApiTestContext):
        # Return a mock config for testing
        return DagsterPlusCliConfig(
            organization=ctx.obj.organization,
            default_deployment=ctx.obj.deployment,
            user_token="test-token",  # Mock token for testing
        )

    # Normal operation - use existing authentication logic
    return get_config_or_error()



import json
from collections.abc import Iterator, Mapping
from pathlib import Path
from typing import Any

import pytest
from click.testing import CliRunner
from dagster_dg_cli.cli.api.client import DgApiTestContext
from dagster_dg_cli.utils.plus.gql_client import DagsterPlusGraphQLClient

from dagster_dg_cli_tests.cli_tests.api_tests.shared.yaml_loader import (
    load_fixture_scenarios_from_yaml,
)


class ReplayClient(DagsterPlusGraphQLClient):
    """GraphQL client that replays recorded responses."""

    def __init__(self, responses: list[dict[str, Any]]):
        # Don't call super().__init__ as we don't need the real GraphQL client
        self.responses = responses
        self.call_index = 0

    def execute(self, query: str, variables: Mapping[str, Any] | None = None) -> dict:
        """Return next recorded response."""
        if self.call_index >= len(self.responses):
            raise ValueError(f"Exhausted {len(self.responses)} responses")
        response = self.responses[self.call_index]
        self.call_index += 1
        return response


def discover_scenario_recordings() -> Iterator[tuple[str, str, str, bool]]:
    """Discover all scenario recordings across API test domains."""
    api_tests_dir = Path(__file__).parent

    domain_dirs = [d for d in api_tests_dir.iterdir() if d.is_dir() and d.name.endswith("_tests")]

    for domain_dir in domain_dirs:
        scenarios_yaml = domain_dir / "scenarios.yaml"
        if not scenarios_yaml.exists():
            continue

        domain = domain_dir.name.replace("_tests", "")
        recording_scenarios = load_fixture_scenarios_from_yaml(scenarios_yaml)

        for scenario_name, scenario_config in recording_scenarios.items():
            yield (domain, scenario_name, scenario_config.command, scenario_config.has_recording)


def load_recorded_graphql_responses(domain: str, scenario_name: str) -> list[dict[str, Any]]:
    """Load GraphQL response recordings for a given domain and scenario."""
    scenario_folder = Path(__file__).parent / f"{domain}_tests" / "recordings" / scenario_name

    if not scenario_folder.exists():
        raise ValueError(f"Recording scenario not found: {scenario_folder}")

    json_files = sorted([f for f in scenario_folder.glob("*.json") if f.name[0:2].isdigit()])

    if not json_files:
        raise ValueError(f"No numbered JSON files found in {scenario_folder}")

    responses = []
    for json_file in json_files:
        with open(json_file) as f:
            responses.append(json.load(f))

    return responses


class TestDynamicCommandExecution:
    """Test all commands from YAML scenarios against recorded GraphQL responses."""

    @pytest.mark.parametrize(
        "domain,scenario_name,command,has_recording", list(discover_scenario_recordings())
    )
    def test_command_execution(
        self, domain: str, scenario_name: str, command: str, has_recording: bool, snapshot
    ):
        """Test executing a command against its recorded GraphQL responses.

        Uses Click's dependency injection to provide a test client factory.
        No mocking required - just inject a context with our replay client.
        """
        from dagster_dg_cli.cli import cli as root_cli
        from dagster_shared.utils.timing import fixed_timezone

        # Load GraphQL responses for this scenario
        graphql_responses = (
            load_recorded_graphql_responses(domain, scenario_name) if has_recording else []
        )

        # Create replay client
        replay_client = ReplayClient(graphql_responses)

        # Create context with test factory
        test_context = DgApiTestContext(client_factory=lambda config: replay_client)

        # Run command with injected context - SINGLE INJECTION POINT
        runner = CliRunner()
        args = command.split()[1:]  # Skip 'dg'

        # This is necessary because some commands output formatted timestamps that don't include
        # timezone info. Fixing the timezone ensures consistent output across environments.
        with fixed_timezone("UTC"):
            # Click's obj parameter passes our context through
            result = runner.invoke(root_cli, args, obj=test_context)

        # Snapshot testing
        output_type = "json" if "--json" in command else "text"
        snapshot_name = f"{domain}_{scenario_name}_{output_type}"

        if "--json" in command:
            try:
                parsed_output = json.loads(result.output)
                assert parsed_output == snapshot(name=snapshot_name)
            except json.JSONDecodeError:
                assert result.output == snapshot(name=snapshot_name)
        else:
            assert result.output == snapshot(name=snapshot_name)

        if "error" not in scenario_name.lower():
            assert result.exit_code == 0



"""GraphQL implementation for deployment operations."""

from typing import TYPE_CHECKING, Any

from dagster_dg_cli.utils.plus.gql_client import IGraphQLClient

if TYPE_CHECKING:
    from dagster_dg_cli.api_layer.schemas.deployment import Deployment, DeploymentList

# GraphQL queries
LIST_DEPLOYMENTS_QUERY = """
query ListDeployments {
    fullDeployments {
        deploymentName
        deploymentId
        deploymentType
    }
}
"""

GET_DEPLOYMENT_QUERY = """
query GetDeployment($deploymentName: String!) {
    deploymentByName(name: $deploymentName) {
        __typename
        ... on DagsterCloudDeployment {
            deploymentName
            deploymentId
            deploymentType
        }
        ... on DeploymentNotFoundError {
            message
        }
        ... on PythonError {
            message
        }
        ... on UnauthorizedError {
            message
        }
    }
}
"""


def process_deployments_response(graphql_response: dict[str, Any]) -> "DeploymentList":
    """Process GraphQL response into DeploymentList.

    This is a pure function that can be easily tested without mocking GraphQL clients.

    Args:
        graphql_response: Raw GraphQL response containing "fullDeployments"

    Returns:
        DeploymentList: Processed deployment data
    """
    # Import pydantic models only when needed
    from dagster_dg_cli.api_layer.schemas.deployment import (
        Deployment,
        DeploymentList,
        DeploymentType,
    )

    deployments_data = graphql_response.get("fullDeployments", [])

    deployments = [
        Deployment(
            id=d["deploymentId"],
            name=d["deploymentName"],
            type=DeploymentType[d["deploymentType"]],
        )
        for d in deployments_data
    ]

    return DeploymentList(
        items=deployments,
        total=len(deployments),
    )


def list_deployments_via_graphql(
    client: IGraphQLClient,
    limit: int | None = None,
) -> "DeploymentList":
    """Fetch deployments using GraphQL.
    This is an implementation detail that can be replaced with REST calls later.
    """
    result = client.execute(LIST_DEPLOYMENTS_QUERY)

    deployment_list = process_deployments_response(result)

    # Apply limit if specified
    if limit is not None:
        deployment_list.items = deployment_list.items[:limit]
        deployment_list.total = len(deployment_list.items)

    return deployment_list


def process_deployment_response(graphql_response: dict[str, Any]) -> "Deployment":
    """Process GraphQL response into single Deployment.

    This is a pure function that can be easily tested without mocking GraphQL clients.

    Args:
        graphql_response: Raw GraphQL response containing "deploymentByName"

    Returns:
        Deployment: Processed deployment data
    """
    # Import pydantic models only when needed
    from dagster_dg_cli.api_layer.schemas.deployment import Deployment, DeploymentType

    deployment_data = graphql_response.get("deploymentByName")

    if not deployment_data:
        raise ValueError("Deployment not found")

    # Handle union type response
    typename = deployment_data.get("__typename")

    if typename == "DagsterCloudDeployment":
        return Deployment(
            id=deployment_data["deploymentId"],
            name=deployment_data["deploymentName"],
            type=DeploymentType[deployment_data["deploymentType"]],
        )
    elif typename == "DeploymentNotFoundError":
        raise ValueError("Deployment not found")
    elif typename in ["PythonError", "UnauthorizedError"]:
        error_message = deployment_data.get("message", "Unknown error")
        raise ValueError(f"Error retrieving deployment: {error_message}")
    else:
        raise ValueError(f"Unexpected response type: {typename}")


def get_deployment_by_name_via_graphql(
    client: IGraphQLClient,
    name: str,
) -> "Deployment":
    """Fetch single deployment by name using GraphQL.
    This is an implementation detail that can be replaced with REST calls later.
    """
    variables = {"deploymentName": name}
    result = client.execute(GET_DEPLOYMENT_QUERY, variables)

    return process_deployment_response(result)



import json
import os
import re
import subprocess
from pathlib import Path

import click
from dagster_dg_core.config import normalize_cli_config
from dagster_dg_core.context import DgContext
from dagster_dg_core.shared_options import dg_global_options
from dagster_dg_core.utils import DgClickCommand, DgClickGroup, exit_with_error
from dagster_dg_core.utils.mcp_client.claude_desktop import get_claude_desktop_config_path
from dagster_dg_core.utils.telemetry import cli_telemetry_wrapper


@click.group(name="mcp", cls=DgClickGroup, hidden=True)
def mcp_group():
    """Commands to interact with the dg MCP server."""


@mcp_group.command(name="serve", cls=DgClickCommand)
@cli_telemetry_wrapper
def serve_command():
    from dagster_dg_cli.mcp.server import mcp

    mcp.run(transport="stdio")


MCP_CONFIG = {
    "command": "dg",
    "args": ["mcp", "serve"],
}


def _inject_into_mcp_config_json(path: Path) -> None:
    contents = json.loads(path.read_text())
    if not contents.get("mcpServers"):
        contents["mcpServers"] = {}
    contents["mcpServers"]["dagster-dg-cli"] = MCP_CONFIG
    path.write_text(json.dumps(contents, indent=2))


def _find_claude() -> str | None:
    try:
        subprocess.run(["claude", "--version"], check=False)
        return "claude"
    except FileNotFoundError:
        pass

    try:  # check for alias (auto-updating version recommends registering an alias instead of putting on PATH)
        result = subprocess.run(
            [os.getenv("SHELL", "bash"), "-ic", "type claude"],
            capture_output=True,
            text=True,
            check=False,
        )
        path_match = re.search(r"(/[^\s`\']+)", result.stdout)
        if path_match:
            return path_match.group(1)
    except FileNotFoundError:
        pass

    return None


@mcp_group.command(name="configure", cls=DgClickCommand)
@dg_global_options
@cli_telemetry_wrapper
@click.argument(
    "mcp_client",
    type=click.Choice(
        [
            "claude-desktop",
            "cursor",
            "claude-code",
            "vscode",
        ]
    ),
)
def configure_mcp_command(
    mcp_client: str,
    **global_options: object,
) -> None:
    """Attempt to add the dg MCP server to the specified MCP client under the key `dagster-dg-cli`."""
    if mcp_client == "claude-desktop":
        _inject_into_mcp_config_json(get_claude_desktop_config_path())
    elif mcp_client == "cursor":
        _inject_into_mcp_config_json(Path.home() / ".cursor" / "mcp.json")
    elif mcp_client == "vscode":
        cli_config = normalize_cli_config(global_options, click.get_current_context())
        dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)
        _inject_into_mcp_config_json(dg_context.root_path / ".vscode" / "mcp.json")
    elif mcp_client == "claude-code":
        claude_cmd = _find_claude()
        if not claude_cmd:
            exit_with_error(
                "Could not find claude code. Please install it from https://github.com/anthropics/claude-code"
            )
        subprocess.run(
            [claude_cmd, "mcp", "remove", "dagster-dg-cli"],
            check=False,
            capture_output=True,
        )
        subprocess.run(
            [claude_cmd, "mcp", "add-json", "dagster-dg-cli", json.dumps(MCP_CONFIG)],
            check=True,
        )



#!/usr/bin/env python3
"""Script to regenerate the YAML test data files for component inspection.

This script generates the expected_schema and expected_example files
that are used to test the YAML generation functionality.
"""

from pathlib import Path

# Import the YAML generation functions from CLI
from dagster_dg_cli.utils.yaml_template_generator import (
    generate_defs_yaml_example_values,
    generate_defs_yaml_schema,
)

# Import the component classes directly
from dagster_test.components import SimplePipesScriptComponent

# Test data directory
TEST_DATA_DIR = Path(__file__).parent / "test_data"

# Component test cases - same as in the test file
TEST_COMPONENTS = [
    {
        "name": "simple_example",
        "component_class": SimplePipesScriptComponent,
        "test_data_dir": "simple_example",
    },
    # Add more components here as they become available in the test environment
]


def get_component_schema(component_class: type) -> dict:
    """Get the JSON schema for a component class."""
    model_cls = component_class.get_model_cls()
    if model_cls is None:
        return {}
    return model_cls.model_json_schema()


def get_component_type_str(component_class: type) -> str:
    """Get the component type string for a component class."""
    return f"{component_class.__module__}.{component_class.__name__}"


def regenerate_test_data():
    """Regenerate all test data files."""
    print("Regenerating YAML test data files...")  # noqa: T201

    for test_case in TEST_COMPONENTS:
        component_class = test_case["component_class"]
        test_data_dir_name = test_case["test_data_dir"]
        component_name = test_case["name"]

        print(f"Processing {component_name}...")  # noqa: T201

        # Get component information
        component_type_str = get_component_type_str(component_class)
        component_schema = get_component_schema(component_class)

        # Generate YAML content
        schema_yaml = generate_defs_yaml_schema(component_type_str, component_schema)
        example_yaml = generate_defs_yaml_example_values(component_type_str, component_schema)

        # Ensure test data directory exists
        test_dir = TEST_DATA_DIR / test_data_dir_name
        test_dir.mkdir(parents=True, exist_ok=True)

        # Write schema file
        schema_file = test_dir / "expected_schema"
        schema_file.write_text(schema_yaml)
        print(f"  Updated {schema_file}")  # noqa: T201

        # Write example file
        example_file = test_dir / "expected_example"
        example_file.write_text(example_yaml)
        print(f"  Updated {example_file}")  # noqa: T201

    print("Done!")  # noqa: T201


if __name__ == "__main__":
    regenerate_test_data()
